"""This is a helper module to simplify code documentation."""

import inspect
from io import StringIO

_PARAMETER_MAPPING = {
    "store": """
store: Callable or str or minimalkv.KeyValueStore
    The store where we can find or store the dataset.

    Can be either ``minimalkv.KeyValueStore``, a minimalkv store url or a
    generic Callable producing a ``minimalkv.KeyValueStore``""",
    "overwrite": """
overwrite: Optional[bool]
    If True, allow overwrite of an existing dataset.""",
    "label_merger": """
label_merger: Optional[Callable]
    By default the shorter label of either the left or right partition is chosen
    as the merged partition label. Supplying a callable here, allows you to override
    the default behavior and create a new label from all input labels
    (depending on the matches this might be more than two values)""",
    "metadata_merger": """
metadata_merger: Optional[Callable]
     By default partition metadata is combined using the :func:`~plateau.io_components.utils.combine_metadata` function.
     You can supply a callable here that implements a custom merge operation on the metadata dictionaries
     (depending on the matches this might be more than two values).""",
    "table": """
table: Optional[str]
    The table to be loaded. If none is specified, the default 'table' is used.""",
    "table_name": """
table_name:
    The table name of the dataset to be loaded. This creates a namespace for
    the partitioning like

    `dataset_uuid/table_name/*`

    This is to support legacy workflows. We recommend not to use this and use the default wherever possible.""",
    "schema": """
schema: SchemaWrapper
    The dataset table schema""",
    "columns": """
columns
    A subset of columns to be loaded.""",
    "dispatch_by": """
dispatch_by: Optional[List[str]]
    List of index columns to group and partition the jobs by.
    There will be one job created for every observed index value
    combination. This may result in either many very small partitions or in
    few very large partitions, depending on the index you are using this on.

    .. admonition:: Secondary indices

        This is also useable in combination with secondary indices where
        the physical file layout may not be aligned with the logically
        requested layout. For optimal performance it is recommended to use
        this for columns which can benefit from predicate pushdown since
        the jobs will fetch their data individually and will *not* shuffle
        data in memory / over network.""",
    "df_serializer": """
df_serializer : Optional[plateau.serialization.DataFrameSerializer]
    A pandas DataFrame serialiser from `plateau.serialization`""",
    "output_dataset_uuid": """
output_dataset_uuid: Optional[str]
    UUID of the newly created dataset""",
    "output_dataset_metadata": """
output_dataset_metadata: Optional[Dict]
    Metadata for the merged target dataset. Will be updated with a
    `merge_datasets__pipeline` key that contains the source dataset uuids for
    the merge.""",
    "output_store": """
output_store : Union[Callable, str, minimalkv.KeyValueStore]
    If given, the resulting dataset is written to this store. By default
    the input store.

    Can be either `minimalkv.KeyValueStore`, a minimalkv store url or a
    generic Callable producing a ``minimalkv.KeyValueStore``""",
    "metadata": """
metadata : Optional[Dict]
    A dictionary used to update the dataset metadata.""",
    "dataset_uuid": """
dataset_uuid: str
    The dataset UUID""",
    "metadata_version": """
metadata_version: Optional[int]
    The dataset metadata version""",
    "partition_on": """
partition_on: List
    Column names by which the dataset should be partitioned by physically.
    These columns may later on be used as an Index to improve query performance.
    Partition columns need to be present in all dataset tables.
    Sensitive to ordering.""",
    "predicate_pushdown_to_io": """
predicate_pushdown_to_io: bool
    Push predicates through to the I/O layer, default True. Disable
    this if you see problems with predicate pushdown for the given
    file even if the file format supports it. Note that this option
    only hides problems in the storage layer that need to be addressed
    there.""",
    "delete_scope": """
delete_scope: List[Dict]
    This defines which partitions are replaced with the input and therefore
    get deleted. It is a lists of query filters for the dataframe in the
    form of a dictionary, e.g.: `[{'column_1': 'value_1'}, {'column_1': 'value_2'}].
    Each query filter will be given to: func: `dataset.query` and the returned
    partitions will be deleted. If no scope is given nothing will be deleted.
    For `plateau.io.dask.update.update_dataset.*` a delayed object resolving to
    a list of dicts is also accepted.""",
    "categoricals": """
categoricals
    Load the provided subset of columns as a :class:`pandas.Categorical`.""",
    "dates_as_object": """
dates_as_object: bool
    Load pyarrow.date{32,64} columns as ``object`` columns in Pandas
    instead of using ``np.datetime64`` to preserve their type. While
    this improves type-safety, this comes at a performance cost.""",
    "predicates": """
predicates: List[List[Tuple[str, str, Any]]
    Optional list of predicates, like `[[('x', '>', 0), ...]`, that are used
    to filter the resulting DataFrame, possibly using predicate pushdown,
    if supported by the file format.
    This parameter is not compatible with filter_query.

    Predicates are expressed in disjunctive normal form (DNF). This means
    that the innermost tuple describes a single column predicate. These
    inner predicates are all combined with a conjunction (AND) into a
    larger predicate. The most outer list then combines all predicates
    with a disjunction (OR). By this, we should be able to express all
    kinds of predicates that are possible using boolean logic.

    Available operators are: `==`, `!=`, `<=`, `>=`, `<`, `>` and `in`.

    Filtering for missings is supported with operators `==`, `!=` and
    `in` and values `np.nan` and `None` for float and string columns
    respectively.

    .. admonition:: Categorical data

        When using order sensitive operators on categorical data we will
        assume that the categories obey a lexicographical ordering.
        This filtering may result in less than optimal performance and may
        be slower than the evaluation on non-categorical data.

    See also :ref:`predicate_pushdown` and :ref:`efficient_querying`""",
    "secondary_indices": """
secondary_indices: List[str]
    A list of columns for which a secondary index should be calculated.""",
    "sort_partitions_by": """
sort_partitions_by: str
    Provide a column after which the data should be sorted before storage to enable predicate pushdown.""",
    "factory": """
factory: plateau.core.factory.DatasetFactory
    A DatasetFactory holding the store and UUID to the source dataset.""",
    "partition_size": """
partition_size: Optional[int]
    Dask bag partition size. Use a larger numbers to decrease scheduler load and overhead, use smaller numbers for a
    fine-grained scheduling and better resilience against worker errors.""",
    "metadata_storage_format": """
metadata_storage_format: str
    Optional list of datastorage format to use. Currently supported is `.json` & `.msgpack.zstd"`""",
    "df_generator": """
df_generator: Iterable[Union[pandas.DataFrame, Dict[str, pandas.DataFrame]]]
    The dataframe(s) to be stored""",
    "default_metadata_version": """
default_metadata_version: int
    Default metadata version. (Note: Metadata version greater than 3 are only supported)""",
    "delayed_tasks": """
delayed_tasks
    A list of delayed objects where each element returns a :class:`pandas.DataFrame`.""",
    "load_dataset_metadata": """
load_dataset_metadata: bool
    Optional argument on whether to load the metadata or not""",
    "dispatch_metadata": """
dispatch_metadata:
    If True, attach dataset user metadata and dataset index information to
    the MetaPartition instances generated.
    Note: This feature is deprecated and this feature toggle is only
    introduced to allow for easier transition.""",
}


def default_docs(func):
    """A decorator which automatically takes care of default parameter
    documentation for common pipeline factory parameters."""
    # TODO (Kshitij68) Bug: The parameters are not come in the same order as listed in the function. For example in `store_dataframes_as_dataset`
    docs = func.__doc__
    new_docs = ""
    signature = inspect.signature(func)

    try:
        buf = StringIO(docs)
        line = "<intentionally non-empty>"
        while line:
            line = buf.readline()

            if "Parameters" in line:
                indentation_level = len(line) - len(line.lstrip())
                artificial_param_docs = [line, buf.readline()]
                # Include the `-----` line
                for param in signature.parameters.keys():
                    doc = _PARAMETER_MAPPING.get(param, None)
                    if doc and param + ":" not in docs:
                        if not doc.endswith("\n"):
                            doc += "\n"
                        if doc.startswith("\n"):
                            doc = doc[1:]
                        doc_indentation_level = len(doc) - len(doc.lstrip())
                        whitespaces_to_add = indentation_level - doc_indentation_level
                        if whitespaces_to_add < 0:
                            raise RuntimeError(
                                f"Indentation detection went wrong for parameter {param}"
                            )
                        # Adjust the indentation dynamically
                        whitespaces = " " * whitespaces_to_add
                        doc = whitespaces + doc
                        doc = doc.replace("\n", "\n" + whitespaces).rstrip() + "\n"
                        # We are checking if the entire docstring associated with the function is present or not
                        if doc not in docs:
                            artificial_param_docs.append(doc)
                new_docs += "".join(artificial_param_docs)
                continue
            new_docs = "".join([new_docs, line])
        func.__doc__ = new_docs
    except Exception as ex:
        func.__doc__ = str(ex)
    return func



import dask.dataframe as dd
import numpy as np
import pandas as pd
import pandas.testing as pdt
import pytest

from plateau.core.factory import DatasetFactory
from plateau.io.dask._shuffle import _KTK_HASH_BUCKET, _hash_bucket
from plateau.io.dask.dataframe import store_dataset_from_ddf, update_dataset_from_ddf
from plateau.io.iter import read_dataset_as_dataframes__iterator


@pytest.mark.parametrize("col", ["range", "range_duplicated", "random"])
def test_hash_bucket(col, num_buckets=5):
    df = pd.DataFrame(
        {
            "range": np.arange(10),
            "range_duplicated": np.repeat(np.arange(2), 5),
            "random": np.random.randint(0, 100, 10),
        }
    )
    hashed = _hash_bucket(df, [col], num_buckets)
    assert (hashed.groupby(col).agg({_KTK_HASH_BUCKET: "nunique"}) == 1).all().all()

    # Check that hashing is consistent for small dataframe sizes (where df.col.nunique() < num_buckets)
    df_sample = df.iloc[[0, 7]]
    hashed_sample = _hash_bucket(df_sample, [col], num_buckets)
    expected = hashed.loc[df_sample.index]
    pdt.assert_frame_equal(expected, hashed_sample)


def test_hashing_determinism():
    """Make sure that the hashing algorithm used by pandas is independent of
    any context variables."""
    df = pd.DataFrame({"range": np.arange(10)})
    hashed = _hash_bucket(df, ["range"], 5)
    expected = pd.DataFrame(
        {
            "range": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            _KTK_HASH_BUCKET: np.uint8([0, 0, 1, 2, 0, 3, 2, 0, 1, 4]),
        }
    )
    pdt.assert_frame_equal(hashed, expected)


@pytest.mark.parametrize("bucket_by", [None, "range"])
def test_update_shuffle_no_partition_on(store_factory, bucket_by):
    df = pd.DataFrame(
        {
            "range": np.arange(10),
            "range_duplicated": np.repeat(np.arange(2), 5),
            "random": np.random.randint(0, 100, 10),
        }
    )
    ddf = dd.from_pandas(df, npartitions=10)

    with pytest.raises(
        ValueError, match="``num_buckets`` must not be None when shuffling data."
    ):
        update_dataset_from_ddf(
            ddf,
            store_factory,
            dataset_uuid="output_dataset_uuid",
            table="table",
            shuffle=True,
            num_buckets=None,
            bucket_by=bucket_by,
        ).compute()

    res_default = update_dataset_from_ddf(
        ddf,
        store_factory,
        dataset_uuid="output_dataset_uuid_default",
        table="table",
        shuffle=True,
        bucket_by=bucket_by,
    ).compute()

    assert len(res_default.partitions) == 1
    res = update_dataset_from_ddf(
        ddf,
        store_factory,
        dataset_uuid="output_dataset_uuid",
        table="table",
        shuffle=True,
        num_buckets=2,
        bucket_by=bucket_by,
    ).compute()

    assert len(res.partitions) == 2


@pytest.mark.parametrize("unique_primaries", [1, 4])
@pytest.mark.parametrize("unique_secondaries", [1, 3])
@pytest.mark.parametrize("num_buckets", [1, 5])
@pytest.mark.parametrize("repartition", [1, 2])
@pytest.mark.parametrize("npartitions", [5, 10])
@pytest.mark.parametrize("bucket_by", [None, "sorted_column"])
@pytest.mark.parametrize("func", [update_dataset_from_ddf, store_dataset_from_ddf])
def test_update_shuffle_buckets(
    store_factory,
    unique_primaries,
    unique_secondaries,
    num_buckets,
    repartition,
    npartitions,
    bucket_by,
    func,
):
    """Assert that certain properties are always given for the output dataset
    no matter how the input data distribution looks like.

    Properties to assert:
    * All partitions have a unique value for its correspondent primary key
    * number of partitions is at least one per unique partition value, at
      most ``num_buckets`` per primary partition value.
    * If we demand a column to be sorted it is per partition monotonic
    """

    primaries = np.arange(unique_primaries)
    secondary = np.arange(unique_secondaries)
    num_rows = 100
    primaries = np.repeat(primaries, np.ceil(num_rows / unique_primaries))[:num_rows]
    secondary = np.repeat(secondary, np.ceil(num_rows / unique_secondaries))[:num_rows]
    # ensure that there is an unsorted column uncorrelated
    # to the primary and secondary columns which can be sorted later on per partition
    unsorted_column = np.repeat(np.arange(100 / 10), 10)
    np.random.shuffle(unsorted_column)
    np.random.shuffle(primaries)
    np.random.shuffle(secondary)

    df = pd.DataFrame(
        {"primary": primaries, "secondary": secondary, "sorted_column": unsorted_column}
    )
    secondary_indices = ["secondary"]
    expected_num_indices = 2  # One primary

    # used for tests later on to
    if bucket_by:
        secondary_indices.append(bucket_by)
        expected_num_indices = 3

    # shuffle all rows. properties of result should be reproducible
    df = df.sample(frac=1).reset_index(drop=True)
    ddf = dd.from_pandas(df, npartitions=npartitions)

    dataset_comp = func(
        ddf,
        store_factory,
        dataset_uuid="output_dataset_uuid",
        table="core",
        secondary_indices=secondary_indices,
        shuffle=True,
        bucket_by=bucket_by,
        repartition_ratio=repartition,
        num_buckets=num_buckets,
        sort_partitions_by="sorted_column",
        partition_on=["primary"],
    )

    dataset = dataset_comp.compute(scheduler="sync")
    dataset = dataset.load_all_indices(store_factory())

    assert len(dataset.partitions) <= num_buckets * unique_primaries
    assert len(dataset.partitions) >= unique_primaries

    assert len(dataset.indices) == expected_num_indices

    assert set(dataset.indices["primary"].index_dct.keys()) == set(
        range(unique_primaries)
    )
    assert [len(x) for x in dataset.indices["primary"].index_dct.values()] <= [
        num_buckets
    ] * unique_primaries

    assert set(dataset.indices["secondary"].index_dct.keys()) == set(
        range(unique_secondaries)
    )

    assert set(dataset.schema.names) == {
        "primary",
        "secondary",
        "sorted_column",
    }

    factory = DatasetFactory("output_dataset_uuid", store_factory)
    factory.load_all_indices()

    if bucket_by:
        ind_df = factory.get_indices_as_dataframe(["primary", bucket_by])

        assert not ind_df.duplicated().any()

    for df in read_dataset_as_dataframes__iterator(
        dataset_uuid=dataset.uuid, store=store_factory
    ):
        assert len(df.primary.unique()) == 1
        assert df.sorted_column.is_monotonic_increasing



from dask.sizeof import sizeof as dask_sizeof


def _dct_sizeof(obj):
    return dask_sizeof(obj.__dict__)


def register_sizeof_ktk_classes():
    from plateau.core.common_metadata import SchemaWrapper
    from plateau.core.dataset import DatasetMetadata
    from plateau.core.factory import DatasetFactory
    from plateau.core.index import ExplicitSecondaryIndex, PartitionIndex
    from plateau.core.partition import Partition
    from plateau.io_components.metapartition import MetaPartition

    dask_sizeof.register(DatasetMetadata, _dct_sizeof)
    dask_sizeof.register(DatasetFactory, _dct_sizeof)
    dask_sizeof.register(MetaPartition, _dct_sizeof)
    dask_sizeof.register(ExplicitSecondaryIndex, _dct_sizeof)
    dask_sizeof.register(PartitionIndex, _dct_sizeof)
    dask_sizeof.register(Partition, _dct_sizeof)
    dask_sizeof.register(SchemaWrapper, _dct_sizeof)



import pandas as pd

from plateau.core.testing import get_dataframe_alltypes


def test_get_dataframe_alltypes():
    df = get_dataframe_alltypes()
    assert isinstance(df, pd.DataFrame)
    assert not df.empty
    assert "byte" in df.columns



from typing import cast

from plateau.core.factory import _ensure_factory
from plateau.core.index import ExplicitSecondaryIndex
from plateau.core.naming import TABLE_METADATA_FILE


def dispatch_files_to_gc(dataset_uuid, store_factory, chunk_size, factory):
    ds_factory = _ensure_factory(
        dataset_uuid=dataset_uuid,
        store=store_factory,
        factory=factory,
    )
    dataset_uuid = dataset_uuid or ds_factory.uuid

    index_path = f"{dataset_uuid}/indices/"
    remove_index_files = set(ds_factory.store.iter_keys(prefix=index_path))

    for index in ds_factory.indices.values():
        index_keys = set()
        # We only add the indices that are saved as explicit indices
        if index.index_storage_key:  # type: ignore
            index_keys.add(cast(ExplicitSecondaryIndex, index).index_storage_key)
        remove_index_files -= index_keys

    remove_table_files = set()
    if ds_factory.explicit_partitions:
        table_files = set()
        for partition in ds_factory.partitions.values():
            for name in partition.files.values():
                table_files.add(name)

        for table in ds_factory.tables:
            table_path = f"{dataset_uuid}/{table}/"
            table_files.add(table_path + TABLE_METADATA_FILE)
            for key in ds_factory.store.iter_keys(prefix=table_path):
                remove_table_files.add(key)
        remove_table_files -= table_files

    files_to_remove = list(remove_index_files | remove_table_files)

    if chunk_size is None:
        yield files_to_remove
    else:
        for i in range(0, len(files_to_remove), chunk_size):
            yield files_to_remove[i : i + chunk_size]


def delete_files(files, store_factory):
    store = store_factory()
    for f in files:
        store.delete(f)



import pickle
from functools import partial

import pytest

from plateau.io.dask.delayed import read_dataset_as_delayed
from plateau.io.testing.read import *  # noqa


@pytest.fixture(params=["table"])
def output_type(request):
    return request.param


def _load_dataframes(output_type, *args, **kwargs):
    if "tables" in kwargs:
        param_tables = kwargs.pop("tables")
        kwargs["table"] = param_tables
    func = partial(read_dataset_as_delayed)
    tasks = func(*args, **kwargs)

    s = pickle.dumps(tasks, pickle.HIGHEST_PROTOCOL)
    tasks = pickle.loads(s)

    result = [task.compute() for task in tasks]
    return result


@pytest.fixture()
def bound_load_dataframes(output_type):
    return partial(_load_dataframes, output_type)



import copy
from typing import TYPE_CHECKING, Any, Optional, TypeVar, cast

from plateau.core.dataset import DatasetMetadata, DatasetMetadataBase
from plateau.core.typing import StoreInput
from plateau.core.utils import lazy_store

if TYPE_CHECKING:
    from minimalkv import KeyValueStore

__all__ = ("DatasetFactory",)

T = TypeVar("T", bound="DatasetFactory")


class DatasetFactory(DatasetMetadataBase):
    """Container holding metadata caching storage access."""

    _nullable_attributes = ["_cache_metadata", "_cache_store"]
    _cache_store: Optional["KeyValueStore"]

    def __init__(
        self,
        dataset_uuid: str,
        store_factory: StoreInput,
        load_schema: bool = True,
        load_all_indices: bool = False,
    ) -> None:
        """A dataset factory object which can be used to cache dataset load
        operations. This class should be the primary user entry point when
        reading datasets.

        Example using the eager backend:

        .. code::

            from functools import partial
            from minimalkv import get_store_from_url
            from plateau.io.eager import read_table

            ds_factory = DatasetFactory(
                dataset_uuid="my_test_dataset",
                store=partial(get_store_from_url, store_url)
            )

            df = read_table(factory=ds_factory)

        Parameters
        ----------
        dataset_uuid
            The unique indetifier for the dataset.
        store_factory
            A callable which creates a KeyValueStore object
        load_schema
            Load the schema information immediately.
        load_all_indices
            Load all indices immediately.
        """
        self._cache_metadata: DatasetMetadata | None = None
        self._cache_store = None

        self.store_factory = lazy_store(store_factory)
        self.dataset_uuid = dataset_uuid
        self.load_schema = load_schema
        self._ds_callable = None
        self.is_loaded = False
        self.load_all_indices_flag = load_all_indices

    def __repr__(self):
        return f"<DatasetFactory: uuid={self.dataset_uuid} is_loaded={self.is_loaded}>"

    @property
    def store(self) -> "KeyValueStore":
        if self._cache_store is None:
            self._cache_store = self.store_factory()
        return self._cache_store

    def _instantiate_metadata_cache(self: T) -> T:
        if self._cache_metadata is None:
            if self._ds_callable:
                # backwards compat
                self._cache_metadata = self._ds_callable()
            else:
                self._cache_metadata = DatasetMetadata.load_from_store(
                    uuid=self.dataset_uuid,
                    store=self.store,
                    load_schema=self.load_schema,
                    load_all_indices=self.load_all_indices_flag,
                )
        self.is_loaded = True
        return self

    @property
    def dataset_metadata(self) -> DatasetMetadata:
        self._instantiate_metadata_cache()
        # The above line ensures non-None
        return cast(DatasetMetadata, self._cache_metadata)

    def invalidate(self) -> None:
        self.is_loaded = False
        self._cache_metadata = None
        self._cache_store = None

    def __getattr__(self, name):
        # __getattr__ should only be called if the attribute cannot be found. if the
        # attribute is None, it still falls back to this call
        if name in self._nullable_attributes:
            return object.__getattribute__(self, name)
        self._instantiate_metadata_cache()
        ds = self.dataset_metadata
        return getattr(ds, name)

    def __getstate__(self):
        # remove cache
        return {k: v for k, v in self.__dict__.items() if not k.startswith("_cache_")}

    def __setstate__(self, state):
        self.__init__(  # type: ignore
            dataset_uuid=state["dataset_uuid"],
            store_factory=state["store_factory"],
            load_schema=state["load_schema"],
            load_all_indices=state["load_all_indices_flag"],
        )

    def __deepcopy__(self, memo) -> "DatasetFactory":
        new_obj = DatasetFactory(
            dataset_uuid=self.dataset_uuid,
            store_factory=self.store_factory,
            load_schema=self.load_schema,
            load_all_indices=self.load_all_indices_flag,
        )
        if self._cache_metadata is not None:
            new_obj._cache_metadata = copy.deepcopy(self._cache_metadata)
        return new_obj

    def load_index(self: T, column, store=None) -> T:
        self._cache_metadata = self.dataset_metadata.load_index(column, self.store)
        return self

    def load_all_indices(self: T, store: Any = None) -> T:
        self._cache_metadata = self.dataset_metadata.load_all_indices(self.store)
        return self

    def load_partition_indices(self: T) -> T:
        self._cache_metadata = self.dataset_metadata.load_partition_indices()
        return self


def _ensure_factory(
    dataset_uuid: str | None,
    store: StoreInput | None,
    factory: DatasetFactory | None,
    load_schema: bool = True,
) -> DatasetFactory:
    if store is None and dataset_uuid is None and factory is not None:
        return factory
    elif store is not None and dataset_uuid is not None and factory is None:
        return DatasetFactory(
            dataset_uuid=dataset_uuid,
            store_factory=lazy_store(store),
            load_schema=load_schema,
        )

    else:
        raise ValueError(
            "Need to supply either a `factory` or `dataset_uuid` and `store`"
        )



import pytest
import simplejson

import plateau.core._zmsgpack as msgpack
from plateau.core.dataset import DatasetMetadata, DatasetMetadataBuilder
from plateau.core.index import ExplicitSecondaryIndex
from plateau.core.partition import Partition
from plateau.core.testing import TIME_TO_FREEZE_ISO


@pytest.mark.parametrize("explicit_partitions", [True, False])
def test_builder_empty(explicit_partitions, metadata_version, frozen_time):
    creation_time = TIME_TO_FREEZE_ISO
    expected = {
        "dataset_uuid": "uuid",
        "dataset_metadata_version": metadata_version,
        "metadata": {"creation_time": creation_time},
    }
    if explicit_partitions:
        expected["partitions"] = {}
    key, result = DatasetMetadataBuilder(
        "uuid",
        metadata_version=metadata_version,
        explicit_partitions=explicit_partitions,
    ).to_json()
    result = simplejson.loads(result)
    assert key == "uuid.by-dataset-metadata.json"
    assert result == expected


def test_builder_msgpack(metadata_version, frozen_time):
    creation_time = TIME_TO_FREEZE_ISO
    expected = {
        "dataset_uuid": "uuid",
        "dataset_metadata_version": metadata_version,
        "metadata": {"creation_time": creation_time},
        "partitions": {},
    }
    key, result = DatasetMetadataBuilder(
        "uuid", metadata_version=metadata_version
    ).to_msgpack()
    result = msgpack.unpackb(result)
    assert key == "uuid.by-dataset-metadata.msgpack.zstd"
    assert result == expected


def test_builder_to_dataset(metadata_version, frozen_time):
    expected = {
        "dataset_uuid": "uuid",
        "dataset_metadata_version": metadata_version,
        "partitions": {"part_2": {"files": {"core": "uuid/core/part_2.parquet"}}},
        "metadata": {"key": "value", "creation_time": TIME_TO_FREEZE_ISO},
        "indices": {"col1": {"a": ["part1"], "b": ["part2"]}},
    }

    builder = DatasetMetadataBuilder("uuid", metadata_version=metadata_version)
    part_2 = Partition("part_2", {"core": "uuid/core/part_2.parquet"})
    builder.add_partition("part_2", part_2)
    builder.add_metadata("key", "value")
    builder.add_embedded_index(
        "col1", ExplicitSecondaryIndex("col1", {"a": ["part1"], "b": ["part2"]})
    )

    result = builder.to_dataset()
    expected_from_dict = DatasetMetadata.from_dict(expected)
    assert result == expected_from_dict


def test_builder_full(metadata_version, frozen_time):
    expected = {
        "dataset_uuid": "uuid",
        "dataset_metadata_version": metadata_version,
        "partitions": {
            "run_id=1/L=1/P=1/part_1": {
                "files": {"core": "uuid/core/run_id=1/L=1/P=1/part_1.parquet"}
            }
        },
        "metadata": {"key": "value", "creation_time": TIME_TO_FREEZE_ISO},
        "indices": {
            "col1": {
                "a": ["run_id=1/L=1/P=1/part_1"],
                "b": ["run_id=2/L=1/P=1/part_1"],
            },
            "col2": "uuid.col2.by-dataset-index.parquet",
        },
        "partition_keys": ["L", "P"],
    }

    builder = DatasetMetadataBuilder(
        "uuid", metadata_version=metadata_version, partition_keys=["L", "P"]
    )
    part_2 = Partition(
        label="run_id=1/L=1/P=1/part_1",
        files={"core": "uuid/core/run_id=1/L=1/P=1/part_1.parquet"},
    )
    builder.add_partition("run_id=1/L=1/P=1/part_1", part_2)
    builder.add_metadata("key", "value")
    builder.add_external_index("col2")
    builder.add_embedded_index(
        "col1",
        ExplicitSecondaryIndex(
            "col1", {"a": ["run_id=1/L=1/P=1/part_1"], "b": ["run_id=2/L=1/P=1/part_1"]}
        ),
    )
    key, result = builder.to_json()
    result = simplejson.loads(result)
    assert key == "uuid.by-dataset-metadata.json"
    assert result == expected


def test_builder_empty_partition_keys(store, metadata_version, frozen_time):
    expected = {
        "dataset_uuid": "uuid",
        "dataset_metadata_version": metadata_version,
        "metadata": {"creation_time": TIME_TO_FREEZE_ISO},
        "partition_keys": ["L", "P"],
        "partitions": {},
    }

    builder = DatasetMetadataBuilder(
        "uuid", metadata_version=4, partition_keys=["L", "P"]
    )
    key, result = builder.to_json()
    result = simplejson.loads(result)
    assert key == "uuid.by-dataset-metadata.json"
    assert result == expected
    result_from_dict = DatasetMetadata.load_from_dict(result, store).to_dict()
    assert result_from_dict == expected



import datetime
import logging
import pickle
from itertools import permutations

import numpy as np
import pandas as pd
import pyarrow as pa
import pytest
import pytz
from hypothesis import assume, example, given
from pandas.testing import assert_series_equal

from plateau.core.index import ExplicitSecondaryIndex, IndexBase, merge_indices
from plateau.core.testing import get_numpy_array_strategy


@pytest.mark.parametrize("inplace", [True, False])
def test_index_update(inplace):
    original_index = ExplicitSecondaryIndex(
        column="col", index_dct={1: ["part_1", "part_2"], 3: ["part_3"]}
    )

    new_index = ExplicitSecondaryIndex(
        column="col", index_dct={1: ["part_4"], 4: ["part_4"]}
    )

    updated_index = original_index.update(new_index, inplace=inplace)

    expected_index = ExplicitSecondaryIndex(
        column="col",
        index_dct={1: ["part_2", "part_4", "part_1"], 3: ["part_3"], 4: ["part_4"]},
    )
    assert updated_index == expected_index


@pytest.mark.parametrize("inplace", [True, False])
def test_storage_key_after_update(inplace):
    """Assert that the storage key is not set after mutation of the index
    object."""
    original_index = ExplicitSecondaryIndex(
        column="col",
        index_dct={1: ["part_1", "part_2"], 3: ["part_3"]},
        index_storage_key="storage_key",
    )
    updated_index = original_index.remove_partitions([], inplace=inplace)
    assert updated_index.index_storage_key == "storage_key"
    updated_index = original_index.remove_partitions(["part_1"], inplace=inplace)
    assert updated_index.index_storage_key is None

    original_index = ExplicitSecondaryIndex(
        column="col",
        index_dct={1: ["part_1", "part_2"], 3: ["part_3"]},
        index_storage_key="storage_key",
    )
    updated_index = original_index.remove_values([], inplace=inplace)
    assert updated_index.index_storage_key == "storage_key"

    updated_index = original_index.remove_values([1], inplace=inplace)
    assert updated_index.index_storage_key is None

    original_index = ExplicitSecondaryIndex(
        column="col",
        index_dct={1: ["part_1", "part_2"], 3: ["part_3"]},
        index_storage_key="storage_key",
    )
    updated_index = original_index.copy()
    assert updated_index.index_storage_key == "storage_key"
    updated_index = original_index.copy(column="something_different")
    assert updated_index.index_storage_key is None


def test_eq_explicit():
    def assert_eq(a, b):
        assert a == b
        assert b == a
        assert not (a != b)
        assert not (b != a)

    def assert_ne(a, b):
        assert a != b
        assert b != a
        assert not (a == b)
        assert not (b == a)

    original_index = ExplicitSecondaryIndex(
        column="col",
        index_dct={1: ["part_1"]},
        dtype=pa.int64(),
        index_storage_key="dataset_uuid/some_index.parquet",
    )

    idx1 = original_index.copy()
    assert_eq(idx1, original_index)

    idx2 = original_index.copy()
    idx2.column = "col2"
    assert_ne(idx2, original_index)

    idx3 = original_index.copy()
    idx3.dtype = pa.uint64()
    assert_ne(idx3, original_index)

    idx4 = original_index.copy()
    idx4.index_dct = {1: ["part_1"], 2: ["part_2"]}
    assert_ne(idx4, original_index)

    idx5 = original_index.copy()
    idx5.index_dct = {1: ["part_1", "part_2"]}
    assert_ne(idx5, original_index)

    idx6 = original_index.copy()
    idx6.index_dct = {1: ["part_2"]}
    assert_ne(idx6, original_index)

    idx7 = original_index.copy()
    idx7.index_dct = {2: ["part_1"]}
    assert_ne(idx7, original_index)

    idx8 = original_index.copy()
    idx8.dtype = None
    assert_ne(idx8, original_index)

    idx9a = original_index.copy()
    idx9b = original_index.copy()
    idx9a.dtype = None
    idx9b.dtype = None
    assert_eq(idx9a, idx9b)


def test_index_update_wrong_col():
    original_index = ExplicitSecondaryIndex(column="col", index_dct={1: ["part_1"]})

    new_index = ExplicitSecondaryIndex(column="another_col", index_dct={1: ["part_4"]})
    with pytest.raises(ValueError) as e:
        original_index.update(new_index)
    assert (
        str(e.value)
        == "Trying to update an index with the wrong column. Got `another_col` but expected `col`"
    )


@pytest.mark.parametrize(
    "dtype",
    [
        pa.binary(),
        pa.bool_(),
        pa.date32(),
        pa.float32(),
        pa.float64(),
        pa.int64(),
        pa.int8(),
        pa.string(),
        pa.timestamp("ns"),
    ],
)
def test_index_empty(store, dtype):
    storage_key = "dataset_uuid/some_index.parquet"
    index1 = ExplicitSecondaryIndex(
        column="col", index_dct={}, dtype=dtype, index_storage_key=storage_key
    )
    key1 = index1.store(store, "dataset_uuid")

    index2 = ExplicitSecondaryIndex(column="col", index_storage_key=key1).load(store)
    assert index1 == index2

    index3 = pickle.loads(pickle.dumps(index1))
    assert index1 == index3


def test_pickle_without_load(store):
    storage_key = "dataset_uuid/some_index.parquet"
    index1 = ExplicitSecondaryIndex(
        column="col", index_dct={1: ["part_1"]}, index_storage_key=storage_key
    )
    key1 = index1.store(store, "dataset_uuid")

    index2 = ExplicitSecondaryIndex(column="col", index_storage_key=key1)
    assert index2 != index1

    index3 = pickle.loads(pickle.dumps(index2))
    assert index3 == index2

    index4 = index3.load(store)
    assert index4 == index1
    assert index4 != index2


def test_index_no_source():
    with pytest.raises(ValueError) as e:
        ExplicitSecondaryIndex(column="col")
    assert str(e.value) == "No valid index source specified"


@pytest.mark.parametrize("inplace", [True, False])
def test_index_remove_values(inplace):
    original_index = ExplicitSecondaryIndex(
        column="col", index_dct={1: ["part_1", "part_2"], 2: ["part_1"], 3: ["part_3"]}
    )
    new_index = original_index.remove_values([1, 2], inplace=inplace)
    expected_index = ExplicitSecondaryIndex(column="col", index_dct={3: ["part_3"]})
    assert new_index == expected_index


@pytest.mark.parametrize("inplace", [True, False])
@pytest.mark.parametrize(
    "remove, expected",
    [
        (["part_1"], {1: ["part_2"], 3: ["part_3"]}),
        (["part_1", "part_2"], {3: ["part_3"]}),
        (["part_1", "part_2", "part_3"], {}),
    ],
)
def test_index_remove_partitions(inplace, remove, expected):
    original_index = ExplicitSecondaryIndex(
        column="col", index_dct={1: ["part_1", "part_2"], 2: ["part_1"], 3: ["part_3"]}
    )
    for perm in permutations(remove):
        new_index = original_index.remove_partitions(perm, inplace=inplace)
        expected_index = ExplicitSecondaryIndex(
            column="col", index_dct=expected, dtype=pa.int64()
        )
        assert new_index == expected_index


def test_index_store_roundtrip_explicit_key(store):
    storage_key = "dataset_uuid/some_index.parquet"
    index1 = ExplicitSecondaryIndex(
        column="col",
        index_dct={1: ["part_1", "part_2"], 3: ["part_3"]},
        index_storage_key=storage_key,
        dtype=pa.int64(),
    )
    key1 = index1.store(store, "dataset_uuid")

    index2 = ExplicitSecondaryIndex(column="col", index_storage_key=key1).load(store)
    assert index1 == index2
    key2 = index2.store(store, "dataset_uuid")

    index3 = ExplicitSecondaryIndex(column="col", index_storage_key=key2).load(store)
    assert index1 == index3
    assert index2 == index3


@pytest.mark.parametrize("col", ["col", "foo/bar", "foo:bar"])
def test_index_store_roundtrip_implicit_key(store, col):
    index1 = ExplicitSecondaryIndex(
        column=col, index_dct={1: ["part_1", "part_2"], 3: ["part_3"]}, dtype=pa.int64()
    )
    key1 = index1.store(store, "dataset_uuid")
    index1.index_storage_key = key1

    index2 = ExplicitSecondaryIndex(column=col, index_storage_key=key1).load(store)
    assert index1 == index2
    key2 = index2.store(store, "dataset_uuid")

    index3 = ExplicitSecondaryIndex(column=col, index_storage_key=key2).load(store)
    assert index1 == index3
    assert index2 == index3


def test_index_as_flat_series():
    index1 = ExplicitSecondaryIndex(
        column="col",
        index_dct={1: ["part_1", "part_2"], 2: ["part_1"]},
        dtype=pa.int64(),
    )
    ser = index1.as_flat_series()
    expected = pd.Series(
        ["part_1", "part_2", "part_1"],
        index=pd.Index([1, 1, 2], name="col"),
        name="partition",
    )
    assert_series_equal(ser, expected)

    ser_comp = index1.as_flat_series(compact=True)
    expected = pd.Series(
        [["part_1", "part_2"], ["part_1"]],
        index=pd.Index([1, 2], name="col"),
        name="partition",
    )
    assert_series_equal(ser_comp, expected)


def test_index_as_flat_series_single_value():
    index1 = ExplicitSecondaryIndex(
        column="col", index_dct={1: ["part_1", "part_2"]}, dtype=pa.int64()
    )
    ser = index1.as_flat_series()
    expected = pd.Series(
        ["part_1", "part_2"], index=pd.Index([1, 1], name="col"), name="partition"
    )
    assert_series_equal(ser, expected)

    ser_comp = index1.as_flat_series(compact=True)
    expected = pd.Series(
        [["part_1", "part_2"]], index=pd.Index([1], name="col"), name="partition"
    )
    assert_series_equal(ser_comp, expected)


def test_index_as_flat_series_partitions_as_index():
    index1 = ExplicitSecondaryIndex(
        column="col",
        index_dct={1: ["part_1", "part_2"], 2: ["part_1"]},
        dtype=pa.int64(),
    )

    ser = index1.as_flat_series(partitions_as_index=True)
    expected = pd.Series(
        [1, 1, 2],
        index=pd.Index(["part_1", "part_2", "part_1"], name="partition"),
        name="col",
    )
    assert_series_equal(ser, expected)

    ser_comp = index1.as_flat_series(compact=True, partitions_as_index=True)
    expected = pd.Series(
        [[1, 2], [1]],
        index=pd.Index(["part_1", "part_2"], name="partition"),
        name="col",
    )
    assert_series_equal(ser_comp, expected)


def test_index_as_flat_series_highly_degenerated_sym():
    dim = 4
    index1 = ExplicitSecondaryIndex(
        column="col",
        index_dct={k: [f"part_{i}" for i in range(0, dim)] for k in range(0, dim)},
        dtype=pa.int64(),
    )
    ser = index1.as_flat_series()
    expected = pd.Series(
        [f"part_{i}" for i in range(0, dim)] * dim,
        index=pd.Index(
            np.array([[i] * dim for i in range(0, dim)]).ravel(), name="col"
        ),
        name="partition",
    )
    assert_series_equal(ser, expected)


def test_index_as_flat_series_highly_degenerated_asym():
    """Ensure that the generation of the series is not bound by col numbers or
    nans in the matrix."""
    dim = 4
    ind_dct = {k: [f"part_{i}" for i in range(0, dim)] for k in range(0, dim)}
    ind_dct[0] = ["part_1"]
    ind_dct[2] = ["part_2", "part_5"]
    index1 = ExplicitSecondaryIndex(column="col", index_dct=ind_dct, dtype=pa.int64())
    ser = index1.as_flat_series()
    partition = [
        "part_1",
        "part_0",
        "part_1",
        "part_2",
        "part_3",
        "part_2",
        "part_5",
        "part_0",
        "part_1",
        "part_2",
        "part_3",
    ]
    index_values = [0, 1, 1, 1, 1, 2, 2, 3, 3, 3, 3]
    expected = pd.Series(
        partition, index=pd.Index(index_values, name="col", dtype=int), name="partition"
    )
    assert_series_equal(ser, expected)

    ser_inv = index1.as_flat_series(partitions_as_index=True)
    expected_inv = pd.Series(
        index_values, index=pd.Index(partition, name="partition"), name="col"
    )
    assert_series_equal(ser_inv, expected_inv)


@pytest.mark.parametrize(
    "dtype, date_as_object", [(None, True), ("datetime64[ns]", False)]
)
def test_index_as_flat_series_date(dtype, date_as_object):
    index1 = ExplicitSecondaryIndex(
        column="col",
        index_dct={
            datetime.date(2017, 1, 2): ["part_1", "part_2"],
            datetime.date(2018, 2, 3): ["part_1"],
        },
        dtype=pa.date32(),
    )
    ser = index1.as_flat_series(date_as_object=date_as_object)
    ser = ser.sort_index()
    expected = pd.Series(
        ["part_1", "part_2", "part_1"],
        index=pd.Index(
            [
                datetime.date(2017, 1, 2),
                datetime.date(2017, 1, 2),
                datetime.date(2018, 2, 3),
            ],
            dtype=dtype,
            name="col",
        ),
        name="partition",
    )
    assert_series_equal(ser, expected)


@pytest.mark.parametrize(
    "dtype, timestamps",
    [
        (pa.timestamp("ns"), [pd.Timestamp("2017-01-01"), pd.Timestamp("2017-01-02")]),
        (
            pa.timestamp("ns"),
            [
                pd.Timestamp("2017-01-01", tzinfo=pytz.timezone("EST")),
                pd.Timestamp("2017-01-02", tzinfo=pytz.timezone("EST")),
            ],
        ),
    ],
)
def test_index_store_roundtrip_ts(store, dtype, timestamps):
    storage_key = "dataset_uuid/some_index.parquet"
    index1 = ExplicitSecondaryIndex(
        column="col",
        index_dct=dict(
            zip(timestamps, [["part_1", "part_2"], ["part_3"]], strict=False)
        ),
        index_storage_key=storage_key,
        dtype=dtype,
    )
    key1 = index1.store(store, "dataset_uuid")

    index2 = ExplicitSecondaryIndex(column="col", index_storage_key=key1).load(store)
    assert index1 == index2


@pytest.mark.parametrize(
    "dtype,expected", [(pa.int8(), pa.int64()), (pa.uint8(), pa.uint64()), (None, None)]
)
def test_index_normalize_dtype(dtype, expected):
    index = ExplicitSecondaryIndex(
        column="col", dtype=dtype, index_storage_key="dataset_uuid/some_index.parquet"
    )
    assert index.dtype == expected


def test_index_raises_nested_dtype():
    with pytest.raises(NotImplementedError) as exc:
        ExplicitSecondaryIndex(
            column="col",
            dtype=pa.list_(pa.int8()),
            index_storage_key="dataset_uuid/some_index.parquet",
        )
    assert str(exc.value) == "Indices w/ nested types are not supported"


def test_index_raises_null_dtype():
    with pytest.raises(NotImplementedError) as exc:
        ExplicitSecondaryIndex(
            column="col",
            dtype=pa.null(),
            index_storage_key="dataset_uuid/some_index.parquet",
        )
    assert str(exc.value) == "Indices w/ null/NA type are not supported"


@pytest.mark.parametrize(
    "dtype,value",
    [
        (pa.bool_(), True),
        (pa.int64(), 1),
        (pa.float64(), 1.1),
        (pa.binary(), b"x"),
        (pa.string(), "x"),
        (pa.timestamp("ns"), pd.Timestamp("2018-01-01").to_datetime64()),
        (pa.date32(), datetime.date(2018, 1, 1)),
        (
            pa.timestamp("ns", tz=pytz.timezone("Europe/Berlin")),
            pd.Timestamp("2018-01-01", tzinfo=pytz.timezone("Europe/Berlin")),
        ),
    ],
)
def test_observed_values_plain(dtype, value):
    ind = ExplicitSecondaryIndex(
        column="col", dtype=dtype, index_dct={value: ["part_label"]}
    )
    observed = ind.observed_values()
    assert len(observed) == 1
    assert list(observed) == [value]


@pytest.mark.parametrize("date_as_object", [None, True, False])
def test_observed_values_date_as_object(date_as_object):
    value = datetime.date(2020, 1, 1)
    ind = ExplicitSecondaryIndex(
        column="col", dtype=pa.date32(), index_dct={value: ["part_label"]}
    )
    observed = ind.observed_values(date_as_object=date_as_object)
    if date_as_object:
        expected = value
    else:
        expected = pd.Timestamp(value).to_datetime64()
    assert len(observed) == 1
    assert observed[0] == expected


@pytest.mark.parametrize(
    "dtype,value,expected",
    [
        (pa.bool_(), True, True),
        (pa.bool_(), False, False),
        (pa.bool_(), 1, True),
        (pa.bool_(), 0, False),
        (pa.bool_(), "True", True),
        (pa.bool_(), "False", False),
        (pa.bool_(), "true", True),
        (pa.bool_(), "false", False),
        (pa.int64(), 1, 1),
        (pa.int64(), "1", 1),
        (pa.int64(), 1.0, 1),
        (pa.float64(), 1.1, 1.1),
        (pa.float64(), "1.1", 1.1),
        (pa.float64(), 1, 1.0),
        (pa.binary(), "x", b"x"),
        (pa.string(), "x", "x"),
        (pa.string(), "ö", "ö"),
        (pa.string(), 1, "1"),
        (pa.string(), "ö".encode(), "ö"),
        (
            pa.timestamp("ns"),
            pd.Timestamp("2018-01-01"),
            pd.Timestamp("2018-01-01").to_datetime64(),
        ),
        (
            pa.timestamp("ns"),
            pd.Timestamp("2018-01-01").to_datetime64(),
            pd.Timestamp("2018-01-01").to_datetime64(),
        ),
        (pa.timestamp("ns"), "2018-01-01", pd.Timestamp("2018-01-01").to_datetime64()),
        (pa.date32(), "2018-01-01", datetime.date(2018, 1, 1)),
        (
            pa.timestamp("ns", tz=pytz.timezone("Europe/Berlin")),
            pd.Timestamp("2018-01-01", tzinfo=pytz.timezone("Europe/Berlin")),
            pd.Timestamp(
                "2018-01-01", tzinfo=pytz.timezone("Europe/Berlin")
            ).to_datetime64(),
        ),
        (
            pa.timestamp("ns", tz=pytz.timezone("Europe/Berlin")),
            "2018-01-01",  # Naive date, is interpreted as being UTC
            pd.Timestamp("2018-01-01", tzinfo=pytz.timezone("UTC")).to_datetime64(),
        ),
    ],
)
def test_index_normalize_value(dtype, value, expected):
    index = ExplicitSecondaryIndex(
        column="col", dtype=dtype, index_storage_key="dataset_uuid/some_index.parquet"
    )
    actual = index.normalize_value(index.dtype, value)
    assert actual == expected
    assert type(actual) is type(expected)


def test_index_normalize_during_init():
    index = ExplicitSecondaryIndex(
        column="col",
        dtype=pa.int8(),
        index_dct={"1": ["a", "b"], 1: ["a", "c"], 2.0: ["d"]},
    )
    expected = {1: ["a", "b", "c"], 2: ["d"]}
    assert index.index_dct == expected


@pytest.mark.parametrize("collision", [True, False])
def test_index_normalize_during_init_warn_collision(collision, caplog):
    index_dct = {1: ["a", "c"], 2.0: ["d"]}
    if collision:
        index_dct["1"] = ["a", "b"]

    caplog.set_level(logging.DEBUG)
    ExplicitSecondaryIndex(column="col", dtype=pa.int8(), index_dct=index_dct)

    warn = [
        t[2]
        for t in caplog.record_tuples
        if t[0] == "plateau.core.index" and t[1] == logging.WARN
    ]

    if collision:
        assert any(
            msg.startswith(
                "Value normalization for index column col resulted in 1 collision(s)."
            )
            for msg in warn
        )
    else:
        assert not any(
            msg.startswith("Value normalization for index column") for msg in warn
        )


def test_index_normalize_during_query():
    index = ExplicitSecondaryIndex(
        column="col", dtype=pa.int64(), index_dct={1: ["a", "b", "c"], 2: ["d"]}
    )
    assert index.query(1) == ["a", "b", "c"]
    assert index.query(2) == ["d"]
    assert index.query("2") == ["d"]
    assert index.query(1.0) == ["a", "b", "c"]


@pytest.mark.parametrize(
    "op, value, expected",
    [
        ("==", 1, {"b", "c", "e"}),
        ("<=", 1, {"a", "b", "c", "e"}),
        (">=", 1, {"b", "c", "e", "f"}),
        ("<", 1, {"a", "b", "c"}),
        (">", 1, {"f"}),
        ("in", [0, 2], {"a", "b", "c", "f"}),
    ],
)
@given(index_data=get_numpy_array_strategy(unique=True, sort=True, allow_nan=False))
@example(
    index_data=np.array(
        [
            b"",
            b"\x01",
            b"\x02",
            b"\x03",
            b"\x04",
            b"\x05",
            b"\x06",
            b"\x07",
            b"\x08",
            b"\t",
        ],
        dtype="|S1",
    )
)
def test_eval_operators(index_data, op, value, expected):
    index = ExplicitSecondaryIndex(
        column="col",
        index_dct={
            index_data[0]: ["a", "b", "c"],
            index_data[1]: ["b", "c", "e"],
            index_data[2]: ["f"],
        },
    )
    assume(len(index.index_dct) == 3)
    result = index.eval_operator(op, index_data[value])

    assert result == expected


def test_eval_operators_type_safety():
    # gh66
    ind = IndexBase(column="col", index_dct={1234: ["part"]}, dtype=pa.int64())
    with pytest.raises(
        TypeError,
        match=r"Unexpected type for predicate: Column 'col' has pandas type 'int64', "
        r"but predicate value '1234'",
    ):
        ind.eval_operator("==", "1234")
    with pytest.raises(
        TypeError,
        match=r"Unexpected type for predicate: Column 'col' has pandas type 'int64', "
        r"but predicate value 1234.0",
    ):
        ind.eval_operator("==", 1234.0)

    assert ind.eval_operator("==", 1234) == {"part"}


@pytest.mark.parametrize("inplace", [True, False])
def test_index_normalize_remove_values(inplace):
    original_index = ExplicitSecondaryIndex(
        column="col", dtype=pa.int64(), index_dct={1: ["a", "b", "c"], 2: ["d"]}
    )

    new_index1 = original_index.copy().remove_values([1, 3], inplace=inplace)
    expected_index1 = ExplicitSecondaryIndex(
        column="col", dtype=pa.int64(), index_dct={2: ["d"]}
    )
    assert new_index1 == expected_index1

    new_index2 = original_index.copy().remove_values([1.0, 3.0], inplace=inplace)
    expected_index2 = ExplicitSecondaryIndex(
        column="col", dtype=pa.int64(), index_dct={2: ["d"]}
    )
    assert new_index2 == expected_index2

    new_index3 = original_index.copy().remove_values(["1", "3"], inplace=inplace)
    expected_index3 = ExplicitSecondaryIndex(
        column="col", dtype=pa.int64(), index_dct={2: ["d"]}
    )
    assert new_index3 == expected_index3


def test_index_ts_inference(store):
    index = ExplicitSecondaryIndex(
        column="col",
        index_dct={
            pd.Timestamp("2017-01-01"): ["part_1", "part_2"],
            pd.Timestamp("2017-01-02"): ["part_3"],
        },
    )
    assert index.dtype == pa.timestamp("ns")


def _dict_to_index(dct):
    new_dct = {}
    for col in dct:
        new_dct[col] = ExplicitSecondaryIndex(col, dct[col])
    return new_dct


@pytest.mark.parametrize("obj_factory", [dict, _dict_to_index])
def test_merge_indices(obj_factory):
    indices = [
        _dict_to_index({"location": {"Loc1": ["label1"], "Loc2": ["label1"]}}),
        _dict_to_index(
            {
                "location": {"Loc3": ["label2"], "Loc2": ["label2"]},
                "product": {"Product1": ["label2"], "Product2": ["label2"]},
            }
        ),
        _dict_to_index(
            {"location": {"Loc4": ["label3"]}, "product": {"Product1": ["label3"]}}
        ),
    ]
    result = merge_indices(indices)
    for key, value in result.items():
        if isinstance(value, ExplicitSecondaryIndex):
            result[key] = value.to_dict()
            for part_list in result[key].values():
                part_list.sort()

    expected = {
        "location": {
            "Loc1": ["label1"],
            "Loc2": ["label1", "label2"],
            "Loc3": ["label2"],
            "Loc4": ["label3"],
        },
        "product": {"Product1": ["label2", "label3"], "Product2": ["label2"]},
    }
    assert result == expected


def test_index_uint():
    index = ExplicitSecondaryIndex(
        column="col",
        index_dct={
            14671423800646041619: ["part_1", "part_2"],
            np.iinfo(np.uint64).max: ["part_1"],
        },
    )
    assert index.dtype == "uint64"


@pytest.mark.parametrize(
    "key",
    [
        True,  # pa.bool_()
        1,  # pa.int64()
        1.1,  # pa.float64()
        b"x",  # pa.binary()
        "ö",  # pa.string()
        pd.Timestamp("2018-01-01").to_datetime64(),  # pa.timestamp("ns")
        pd.Timestamp(
            "2018-01-01", tzinfo=pytz.timezone("Europe/Berlin")
        ).to_datetime64(),  # pa.timestamp("ns")
        datetime.date(2018, 1, 1),  # pa.date32()
    ],
)
def test_serialization(key):
    """Check index remains consistent after serializing and de-serializing."""
    index = ExplicitSecondaryIndex(
        column="col", index_dct={key: ["part_2", "part_4", "part_1"]}
    )
    index2 = pickle.loads(pickle.dumps(index))

    assert index == index2


@pytest.mark.parametrize(
    "key",
    [
        True,  # pa.bool_()
        1,  # pa.int64()
        1.1,  # pa.float64()
        b"x",  # pa.binary()
        "ö",  # pa.string()
        pd.Timestamp("2018-01-01").to_datetime64(),  # pa.timestamp("ns")
        pd.Timestamp(
            "2018-01-01", tzinfo=pytz.timezone("Europe/Berlin")
        ).to_datetime64(),  # pa.timestamp("ns")
        datetime.datetime(
            2018, 1, 1, 12, 30
        ),  # pa.timestamp("us") (initial) => pa.timestamp("ns") (after loading)
        datetime.datetime(
            2018, 1, 1, 12, 30, tzinfo=pytz.timezone("Europe/Berlin")
        ),  # pa.timestamp("ns")
        datetime.date(2018, 1, 1),  # pa.date32()
    ],
)
def test_serialization_normalization(key):
    """Check that index normalizes values consistently after serializing.

    This is helpful to ensure correct behavior for cases such as when
    key=`datetime.datetime(2018, 1, 1, 12, 30)`, as this would be parsed
    to `pa.timestamp("us")` during index creation, but stored as
    `pa.timestamp("ns")`.
    """
    index = ExplicitSecondaryIndex(
        column="col", index_dct={key: ["part_2", "part_4", "part_1"]}
    )
    index2 = pickle.loads(pickle.dumps(index))

    assert index.normalize_value(index.dtype, key) == index2.normalize_value(
        index2.dtype, key
    )


@pytest.mark.parametrize("with_index_dct", [True, False])
def test_unload(with_index_dct):
    storage_key = "dataset_uuid/some_index.parquet"
    index1 = ExplicitSecondaryIndex(
        column="col",
        index_dct={1: ["part_1"]} if with_index_dct else None,
        index_storage_key=storage_key,
    )

    index2 = index1.unload()
    assert not index2.loaded
    assert index2.index_storage_key == storage_key

    index3 = pickle.loads(pickle.dumps(index2))
    assert index2 == index3


def test_fail_type_unsafe():
    with pytest.raises(ValueError, match="Trying to create non-typesafe index"):
        ExplicitSecondaryIndex(column="col", index_dct={})


def test_index_large(store):
    storage_key = "dataset_uuid/some_index.parquet"
    index1 = ExplicitSecondaryIndex(
        column="col",
        index_dct={i: ["part_1"] for i in range(100_000)},
        index_storage_key=storage_key,
    )
    key1 = index1.store(store, "dataset_uuid")

    index2 = ExplicitSecondaryIndex(column="col", index_storage_key=key1).load(store)
    assert index1 == index2

    index3 = pickle.loads(pickle.dumps(index1))
    assert index1 == index3



import pytest

from plateau.serialization import CsvSerializer, ParquetSerializer


@pytest.mark.parametrize(
    "serialiser,expected",
    [
        (CsvSerializer(), "CsvSerializer(compress=True)"),
        (CsvSerializer(compress=False), "CsvSerializer(compress=False)"),
        (
            ParquetSerializer(),
            "ParquetSerializer(compression='SNAPPY', chunk_size=None)",
        ),
        (
            ParquetSerializer(compression="GZIP"),
            "ParquetSerializer(compression='GZIP', chunk_size=None)",
        ),
        (
            ParquetSerializer(chunk_size=1000),
            "ParquetSerializer(compression='SNAPPY', chunk_size=1000)",
        ),
    ],
)
def test_repr(serialiser, expected):
    actual = repr(serialiser)
    assert actual == expected


@pytest.mark.parametrize(
    "obj1,obj2,expected",
    [
        (CsvSerializer(), CsvSerializer(), True),
        (CsvSerializer(), CsvSerializer(compress=False), False),
        (CsvSerializer(), ParquetSerializer(), False),
        (ParquetSerializer(), ParquetSerializer(), True),
        (ParquetSerializer(), ParquetSerializer(compression="GZIP"), False),
        (ParquetSerializer(), ParquetSerializer(chunk_size=1000), False),
        (ParquetSerializer(), CsvSerializer(), False),
    ],
)
def test_eq(obj1, obj2, expected):
    actual = obj1 == obj2
    assert actual == expected

    not_actual = obj1 != obj2
    not_expected = not expected
    assert not_actual == not_expected



import pickle

import pytest

from plateau.io.dask.bag import build_dataset_indices__bag
from plateau.io.testing.index import *  # noqa: F403, F4


def _build_indices(*args, **kwargs):
    bag = build_dataset_indices__bag(*args, **kwargs)

    # pickle roundtrip to ensure we don't need the inefficient cloudpickle fallback
    s = pickle.dumps(bag, pickle.HIGHEST_PROTOCOL)
    bag = pickle.loads(s)

    bag.compute()


@pytest.fixture()
def bound_build_dataset_indices():
    return _build_indices



from functools import partial
from typing import cast

from plateau.core.docs import default_docs
from plateau.core.factory import _ensure_factory
from plateau.core.naming import (
    DEFAULT_METADATA_STORAGE_FORMAT,
    DEFAULT_METADATA_VERSION,
    SINGLE_TABLE,
)
from plateau.core.uuid import gen_uuid
from plateau.io_components.metapartition import (
    MetaPartition,
    parse_input_to_metapartition,
)
from plateau.io_components.read import dispatch_metapartitions_from_factory
from plateau.io_components.update import update_dataset_from_partitions
from plateau.io_components.utils import (
    _ensure_compatible_indices,
    normalize_args,
    raise_if_indices_overlap,
    sort_values_categorical,
    validate_partition_keys,
)
from plateau.io_components.write import (
    raise_if_dataset_exists,
    store_dataset_from_partitions,
)

__all__ = (
    "read_dataset_as_dataframes__iterator",
    "update_dataset_from_dataframes__iter",
    "store_dataframes_as_dataset__iter",
)


@default_docs
@normalize_args
def read_dataset_as_metapartitions__iterator(
    dataset_uuid=None,
    store=None,
    columns=None,
    predicate_pushdown_to_io=True,
    categoricals=None,
    dates_as_object: bool = True,
    predicates=None,
    factory=None,
    dispatch_by=None,
):
    """A Python iterator to retrieve a dataset from store where each partition
    is loaded as a :class:`~plateau.io_components.metapartition.MetaPartition`.

    .. seealso:

        :func:`~plateau.io_components.read.read_dataset_as_dataframes__iterator`

    Parameters
    ----------
    """

    ds_factory = _ensure_factory(
        dataset_uuid=dataset_uuid,
        store=store,
        factory=factory,
    )

    store = ds_factory.store
    mps = dispatch_metapartitions_from_factory(
        ds_factory,
        predicates=predicates,
        dispatch_by=dispatch_by,
    )

    for mp in mps:
        if dispatch_by is not None:
            mp = MetaPartition.concat_metapartitions(
                [
                    mp_inner.load_dataframes(
                        store=store,
                        columns=columns,
                        categoricals=categoricals,
                        predicate_pushdown_to_io=predicate_pushdown_to_io,
                        predicates=predicates,
                    )
                    for mp_inner in mp
                ]
            )
        else:
            mp = cast(MetaPartition, mp)
            mp = mp.load_dataframes(
                store=store,
                columns=columns,
                categoricals=categoricals,
                predicate_pushdown_to_io=predicate_pushdown_to_io,
                dates_as_object=dates_as_object,
                predicates=predicates,
            )
        yield mp


@default_docs
@normalize_args
def read_dataset_as_dataframes__iterator(
    dataset_uuid=None,
    store=None,
    columns=None,
    predicate_pushdown_to_io=True,
    categoricals=None,
    dates_as_object: bool = True,
    predicates=None,
    factory=None,
    dispatch_by=None,
):
    """A Python iterator to retrieve a dataset from store where each partition
    is loaded as a :class:`~pandas.DataFrame`.

    Parameters
    ----------

    Returns
    -------
    list
        A list containing a dictionary for each partition. The dictionaries
        keys are the in-partition file labels and the values are the
        corresponding dataframes.

    Examples
    --------
    Dataset in store contains two partitions with two files each

    .. code ::

        >>> import minimalkv
        >>> from plateau.io.iter import read_dataset_as_dataframes__iterator

        >>> store = minimalkv.get_store_from_url('s3://bucket_with_dataset')

        >>> dataframes = read_dataset_as_dataframes__iterator('dataset_uuid', store)
        >>> next(dataframes)
        [
            # First partition
            {'core': pd.DataFrame, 'lookup': pd.DataFrame}
        ]

        >>> next(dataframes)
        [
            # Second partition
            {'core': pd.DataFrame, 'lookup': pd.DataFrame}
        ]
    """
    mp_iter = read_dataset_as_metapartitions__iterator(
        dataset_uuid=dataset_uuid,
        store=store,
        columns=columns,
        predicate_pushdown_to_io=predicate_pushdown_to_io,
        categoricals=categoricals,
        dates_as_object=dates_as_object,
        predicates=predicates,
        factory=factory,
        dispatch_by=dispatch_by,
    )
    for mp in mp_iter:
        yield mp.data


@default_docs
@normalize_args
def update_dataset_from_dataframes__iter(
    df_generator,
    store=None,
    dataset_uuid=None,
    delete_scope=None,
    metadata=None,
    df_serializer=None,
    metadata_merger=None,
    default_metadata_version=DEFAULT_METADATA_VERSION,
    partition_on=None,
    sort_partitions_by=None,
    secondary_indices=None,
    factory=None,
    table_name: str = SINGLE_TABLE,
):
    """Update a plateau dataset in store iteratively, using a generator of
    dataframes.

    Useful for datasets which do not fit into memory.

    Parameters
    ----------

    Returns
    -------
    The dataset metadata object (:class:`~plateau.core.dataset.DatasetMetadata`).

    See Also
    --------
    :ref:`mutating_datasets`
    """

    ds_factory, metadata_version, partition_on = validate_partition_keys(
        dataset_uuid=dataset_uuid,
        store=store,
        ds_factory=factory,
        default_metadata_version=default_metadata_version,
        partition_on=partition_on,
    )

    secondary_indices = _ensure_compatible_indices(ds_factory, secondary_indices)

    if sort_partitions_by:  # Define function which sorts each partition by column
        sort_partitions_by_fn = partial(
            sort_values_categorical, columns=sort_partitions_by
        )

    new_partitions = []
    for df in df_generator:
        mp = parse_input_to_metapartition(
            df,
            metadata_version=metadata_version,
            table_name=table_name,
        )

        if sort_partitions_by:
            mp = mp.apply(sort_partitions_by_fn)

        if partition_on:
            mp = mp.partition_on(partition_on=partition_on)

        if secondary_indices:
            mp = mp.build_indices(columns=secondary_indices)

        # Store dataframe, thereby clearing up the dataframe from the `mp` metapartition
        mp = mp.store_dataframes(
            store=store, df_serializer=df_serializer, dataset_uuid=dataset_uuid
        )

        new_partitions.append(mp)

    return update_dataset_from_partitions(
        new_partitions,
        store_factory=store,
        dataset_uuid=dataset_uuid,
        ds_factory=ds_factory,
        delete_scope=delete_scope,
        metadata=metadata,
        metadata_merger=metadata_merger,
    )


@default_docs
@normalize_args
def store_dataframes_as_dataset__iter(
    df_generator,
    store,
    dataset_uuid=None,
    metadata=None,
    partition_on=None,
    df_serializer=None,
    overwrite=False,
    metadata_storage_format=DEFAULT_METADATA_STORAGE_FORMAT,
    metadata_version=DEFAULT_METADATA_VERSION,
    secondary_indices=None,
    table_name: str = SINGLE_TABLE,
):
    """Store `pd.DataFrame` s iteratively as a partitioned dataset with
    multiple tables (files).

    Useful for datasets which do not fit into memory.

    Parameters
    ----------

    Returns
    -------
    dataset: plateau.core.dataset.DatasetMetadata
        The stored dataset.
    """

    if dataset_uuid is None:
        dataset_uuid = gen_uuid()

    if not overwrite:
        raise_if_dataset_exists(dataset_uuid=dataset_uuid, store=store)

    raise_if_indices_overlap(partition_on, secondary_indices)

    new_partitions = []
    for df in df_generator:
        mp = parse_input_to_metapartition(
            df, metadata_version=metadata_version, table_name=table_name
        )

        if partition_on:
            mp = mp.partition_on(partition_on)

        if secondary_indices:
            mp = mp.build_indices(secondary_indices)

        # Store dataframe, thereby clearing up the dataframe from the `mp` metapartition
        mp = mp.store_dataframes(
            store=store, dataset_uuid=dataset_uuid, df_serializer=df_serializer
        )

        # Add `plateau.io_components.metapartition.MetaPartition` object to list to track partitions
        new_partitions.append(mp)

    # Store metadata and return `plateau.DatasetMetadata` object
    return store_dataset_from_partitions(
        partition_list=new_partitions,
        dataset_uuid=dataset_uuid,
        store=store,
        dataset_metadata=metadata,
        metadata_storage_format=metadata_storage_format,
    )



from ._csv import CsvSerializer
from ._generic import (
    ConjunctionType,
    DataFrameSerializer,
    LiteralType,
    LiteralValue,
    PredicatesType,
    check_predicates,
    columns_in_predicates,
    filter_array_like,
    filter_df,
    filter_df_from_predicates,
    filter_predicates_by_column,
)
from ._parquet import ParquetSerializer

DataFrameSerializer.register_serializer(".csv.gz", CsvSerializer)
DataFrameSerializer.register_serializer(".csv", CsvSerializer)
DataFrameSerializer.register_serializer(".parquet", ParquetSerializer)


def default_serializer():
    return ParquetSerializer()


__all__ = (
    # Serializer classes
    "CsvSerializer",
    "DataFrameSerializer",
    "default_serializer",
    "ParquetSerializer",
    # functions
    "check_predicates",
    "columns_in_predicates",
    "filter_array_like",
    "filter_df_from_predicates",
    "filter_df",
    "filter_predicates_by_column",
    # types
    "ConjunctionType",
    "LiteralType",
    "LiteralValue",
    "PredicatesType",
)



import pickle

import dask
import pytest

from plateau.io.dask.delayed import update_dataset_from_delayed
from plateau.io.testing.update import *  # noqa


@pytest.fixture
def bound_update_dataset():
    return _update_dataset


@dask.delayed
def _unwrap_partition(part):
    return next(iter(dict(part["data"]).values()))


def _update_dataset(partitions, *args, **kwargs):
    if not isinstance(partitions, list):
        partitions = [partitions]
    tasks = update_dataset_from_delayed(partitions, *args, **kwargs)

    s = pickle.dumps(tasks, pickle.HIGHEST_PROTOCOL)
    tasks = pickle.loads(s)

    return tasks.compute()



from collections.abc import Callable
from functools import partial

import numpy as np
import pandas as pd
import pandas.testing as pdt
import pyarrow as pa
import pytest
from minimalkv import get_store_from_url

from plateau.io_components.utils import (
    align_categories,
    combine_metadata,
    extract_duplicates,
    normalize_args,
    sort_values_categorical,
)


def testcombine_metadata():
    d1 = {"a": 1, "b": 1}
    d2 = {"a": 1, "b": 2}
    d3 = {"a": 2, "b": 2}
    d4 = {"a": 1, "c": 3}
    assert combine_metadata([d1]) == d1
    assert combine_metadata([d1, d1]) == d1
    assert combine_metadata([d1, d2]) == {"a": 1}
    assert combine_metadata([d1, d2, d3]) == {}
    assert combine_metadata([d1, d4]) == {"a": 1, "b": 1, "c": 3}


def testcombine_metadata_incompatible():
    d1 = {"a": 1, "b": [1]}
    d2 = {"a": 1, "b": 1}
    assert combine_metadata([d1, d2]) == {"a": 1}


def testcombine_metadata_nested():
    d1 = {"a": {"l2a": 1}, "b": {"l2b": 1}}
    d2 = {"a": {"l2a": 1}, "b": {"l2b": 2}}
    d3 = {"a": {"l2a": 2}, "b": {"l2b": 2}}
    d4 = {"a": {"l2a": 1}, "c": {"l2c": 3}}
    assert combine_metadata([d1]) == d1
    assert combine_metadata([d1, d1]) == d1
    assert combine_metadata([d1, d2]) == {"a": {"l2a": 1}}
    assert combine_metadata([d1, d2, d3]) == {}
    assert combine_metadata([d1, d4]) == {
        "a": {"l2a": 1},
        "b": {"l2b": 1},
        "c": {"l2c": 3},
    }


def testcombine_metadata_nested_lists():
    d1 = {"a": [[1], [2]], "b": [[1], [2]]}
    assert combine_metadata([d1]) == d1

    assert combine_metadata([d1, d1]) == {"a": [[1], [2]], "b": [[1], [2]]}


def testcombine_metadata_lists():
    d1 = {"a": [1, 2], "b": [1, 2]}
    assert combine_metadata([d1], append_to_list=False) == d1

    combined_metadata = combine_metadata([d1, d1], append_to_list=False)
    assert set(combined_metadata["a"]) == {1, 2}
    assert set(combined_metadata["b"]) == {1, 2}


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (("a", "a", "a"), ("a", ["a"], ["a"])),
        (("a", "abc", "a"), ("a", ["abc"], ["a"])),
        (("a", "a", "abc"), ("a", ["a"], ["abc"])),
        (("a", "abc", "abc"), ("a", ["abc"], ["abc"])),
        (("a", "abc", None), ("a", ["abc"], [])),
        (("a", None, "abc"), ("a", [], ["abc"])),
        (("a", None, None), ("a", [], [])),
        (("a", (1, 2), ("a", "b")), ("a", [1, 2], ["a", "b"])),
    ],
)
def test_normalize_args(test_input, expected):
    @normalize_args
    def func(arg1, partition_on, delete_scope=None):
        return arg1, partition_on, delete_scope

    test_arg1, test_partition_on, test_delete_scope = test_input
    assert expected == func(
        test_arg1, test_partition_on, delete_scope=test_delete_scope
    )
    assert (expected[0], expected[1], []) == func(test_arg1, test_partition_on)


@pytest.mark.parametrize("_type", ["callable", "url", "minimalkv"])
def test_normalize_store(tmpdir, _type):
    store_url = f"hfs://{tmpdir}"
    store = get_store_from_url(store_url)
    store.put("test", b"")

    @normalize_args
    def func(store):
        assert isinstance(store, Callable)
        return store().keys()

    if _type == "callable":
        store_test = partial(get_store_from_url, store_url)
    elif _type == "url":
        store_test = store_url
    elif _type == "minimalkv":
        store_test = store
    else:
        raise AssertionError(f"unknown parametrization {_type}")
    assert func(store_test)


@pytest.mark.parametrize(
    "test_input", [("a", {"a"}, "c"), ("a", frozenset("a"), None), ("abc", {"c": 6}, 4)]
)
def test_normalize_args__incompatible_types(test_input):
    @normalize_args
    def func(arg1, partition_on, delete_scope=None):
        return arg1, partition_on, delete_scope

    test_arg1, test_partition_on, test_delete_scope = test_input
    with pytest.raises(ValueError):
        func(test_arg1, test_partition_on, delete_scope=test_delete_scope)


@pytest.mark.parametrize(
    "lst, expected",
    [
        ([], []),
        (["a"], []),
        (["a", "b", "c"], []),
        (["a", "b", "c", "a"], ["a"]),
        (["a", "b", "c", "a", "a"], ["a"]),
        (["a", "a"], ["a"]),
    ],
)
def test_extract_duplicates(lst, expected):
    assert set(extract_duplicates(lst)) == set(expected)


def test_align_categories():
    df1 = pd.DataFrame(
        {
            "col_A": pd.Categorical(["A1", "A3", "A3"]),
            "col_B": pd.Categorical(["B1", "B3", "B3"]),
        }
    )
    df2 = pd.DataFrame(
        {
            "col_A": pd.Categorical(["A2", "A3", "A4"]),
            "col_B": pd.Categorical(["B2", "B3", "B4"]),
        }
    )
    df3 = pd.DataFrame(
        {
            "col_A": pd.Categorical(["A4", "A5", "A1"]),
            "col_B": pd.Categorical(["B4", "B5", "B1"]),
        }
    )
    in_dfs = [df1, df2, df3]

    out_dfs = align_categories(in_dfs, categoricals=["col_A", "col_B"])

    for prefix in ["A", "B"]:
        col_name = f"col_{prefix}"
        expected_categories = [
            f"{prefix}1",
            f"{prefix}3",
            f"{prefix}2",
            f"{prefix}4",
            f"{prefix}5",
        ]
        expected_1 = pd.Series(
            pd.Categorical(
                [f"{prefix}1", f"{prefix}3", f"{prefix}3"],
                categories=expected_categories,
            ),
            name=col_name,
        )
        pdt.assert_series_equal(out_dfs[0][col_name], expected_1)

        expected_2 = pd.Series(
            pd.Categorical(
                [f"{prefix}2", f"{prefix}3", f"{prefix}4"],
                categories=expected_categories,
            ),
            name=col_name,
        )
        pdt.assert_series_equal(out_dfs[1][col_name], expected_2)

        expected_3 = pd.Series(
            pd.Categorical(
                [f"{prefix}4", f"{prefix}5", f"{prefix}1"],
                categories=expected_categories,
            ),
            name=col_name,
        )
        pdt.assert_series_equal(out_dfs[2][col_name], expected_3)


def test_align_categories_with_missings():
    df_0 = pd.DataFrame({"letters": ["a", "a", "b", np.nan]})
    df_1 = pd.DataFrame({"letters": ["a", "a"]})
    out = align_categories([df_0, df_1], ["letters"])
    expected_0 = pd.DataFrame(
        {"letters": pd.Categorical(["a", "a", "b", np.nan], categories=["a", "b"])}
    )
    expected_1 = pd.DataFrame(
        {"letters": pd.Categorical(["a", "a"], categories=["a", "b"])}
    )
    pdt.assert_frame_equal(out[0], expected_0)
    pdt.assert_frame_equal(out[1], expected_1)


def test_sort_categorical():
    values = ["f", "a", "b", "z", "e"]
    categories = ["e", "z", "b", "a", "f"]
    cat_ser = pd.Series(pd.Categorical(values, categories=categories))
    df = pd.DataFrame({"cat": cat_ser, "int": range(len(values))})

    sorted_df = sort_values_categorical(df, "cat")

    expected_values = sorted(values)

    assert all(sorted_df["cat"].values == expected_values)
    assert sorted_df["cat"].is_monotonic_increasing
    assert sorted_df["cat"].cat.ordered
    assert all(sorted_df["cat"].cat.categories == sorted(categories))


def test_sort_categorical_multiple_cols():
    df = pd.DataFrame.from_records(
        [
            {"ColA": "B", "ColB": "Z", "Payload": 1},
            {"ColA": "B", "ColB": "A", "Payload": 2},
            {"ColA": "A", "ColB": "A", "Payload": 3},
            {"ColA": "C", "ColB": "Z", "Payload": 4},
        ]
    )
    df_expected = df.copy().sort_values(by=["ColA", "ColB"]).reset_index(drop=True)
    df = df.astype(dict.fromkeys(["ColA", "ColB"], "category"))
    # Correct order
    # {"ColA": "A", "ColB": "A", "Payload": 3},
    # {"ColA": "B", "ColB": "A", "Payload": 2},
    # {"ColA": "B", "ColB": "Z", "Payload": 1},
    # {"ColA": "C", "ColB": "Z", "Payload": 4},
    df_expected = df_expected.astype(
        {
            "ColA": pd.CategoricalDtype(categories=["A", "B", "C"], ordered=True),
            "ColB": pd.CategoricalDtype(categories=["A", "Z"], ordered=True),
        }
    )

    sorted_df = sort_values_categorical(df, ["ColA", "ColB"])

    pdt.assert_frame_equal(sorted_df, df_expected)


def test_sort_categorical_pyarrow_conversion():
    """Make sure sorting does not introduce indices that end up in the Arrow
    table."""
    df = pd.DataFrame({"a": [3, 2, 1]})
    sorted_df = sort_values_categorical(df, "a")
    table = pa.Table.from_pandas(df)
    sorted_table = pa.Table.from_pandas(sorted_df)
    assert table.schema.names == sorted_table.schema.names



#!/usr/bin/env python


import datetime

import numpy as np
import pandas as pd
import pandas.testing as pdt
import pyarrow as pa
import pytest
from packaging.version import Version
from pyarrow.parquet import ParquetFile

from plateau.core._compat import pandas_infer_string
from plateau.serialization import (
    CsvSerializer,
    DataFrameSerializer,
    ParquetSerializer,
    default_serializer,
)
from plateau.serialization._util import ensure_unicode_string_type

TYPE_STABLE_SERIALISERS = [ParquetSerializer()]

SERLIALISERS = TYPE_STABLE_SERIALISERS + [
    CsvSerializer(),
    CsvSerializer(compress=False),
    default_serializer(),
]

type_stable_serialisers = pytest.mark.parametrize("serialiser", TYPE_STABLE_SERIALISERS)

predicate_serialisers = pytest.mark.parametrize(
    "serialiser",
    [
        ParquetSerializer(chunk_size=1),
        ParquetSerializer(chunk_size=2),
        ParquetSerializer(chunk_size=4),
    ]
    + SERLIALISERS,
)


def assert_frame_equal_non_strict(x, y, *args, **kwargs):
    # assert_X_equal will be stricter with NaN/None equality in Pandas 3
    # (https://github.com/pandas-dev/pandas/pull/52081). Until the strict_na argument is
    # introduced (https://github.com/pandas-dev/pandas/pull/58072), cast np.nan to None.
    if Version(pd.__version__) < Version("3.0.0.dev0"):
        pdt.assert_frame_equal(x, y, *args, **kwargs)
    else:
        pdt.assert_frame_equal(
            x.replace(np.nan, None), y.replace(np.nan, None), *args, **kwargs
        )


def test_load_df_from_store_unsupported_format(store):
    with pytest.raises(ValueError):
        DataFrameSerializer.restore_dataframe(store, "test.unknown")


def test_store_df_to_store(store):
    df = pd.DataFrame({"a": [1, 2], "b": [3.0, 4.0], "c": ["∆", "€"]})
    dataframe_format = default_serializer()
    assert isinstance(dataframe_format, ParquetSerializer)
    key = dataframe_format.store(store, "prefix", df)
    pdt.assert_frame_equal(DataFrameSerializer.restore_dataframe(store, key), df)


@pytest.mark.parametrize("serialiser", SERLIALISERS)
def test_store_table_to_store(serialiser, store):
    df = pd.DataFrame({"a": [1, 2], "b": [3.0, 4.0], "c": ["∆", "€"]})
    table = pa.Table.from_pandas(df)
    key = serialiser.store(store, "prefix", table)
    pdt.assert_frame_equal(DataFrameSerializer.restore_dataframe(store, key), df)


@pytest.mark.parametrize("serialiser", SERLIALISERS)
def test_dataframe_roundtrip(serialiser, store):
    if serialiser in TYPE_STABLE_SERIALISERS:
        df = pd.DataFrame(
            {"a": [1, 2], "b": [3.0, 4.0], "c": ["∆", "€"], b"d": ["#", ";"]}
        )
        key = serialiser.store(store, "prefix", df)
        df.columns = [ensure_unicode_string_type(col) for col in df.columns]
    else:
        df = pd.DataFrame(
            {"a": [1, 2], "b": [3.0, 4.0], "c": ["∆", "€"], "d": ["#", ";"]}
        )
        key = serialiser.store(store, "prefix", df)

    pdt.assert_frame_equal(DataFrameSerializer.restore_dataframe(store, key), df)

    # Test partial restore
    pdt.assert_frame_equal(
        DataFrameSerializer.restore_dataframe(store, key, columns=["a", "c"]),
        df[["a", "c"]],
    )

    # Test that all serialisers can ingest predicate_pushdown_to_io
    pdt.assert_frame_equal(
        DataFrameSerializer.restore_dataframe(
            store, key, columns=["a", "c"], predicate_pushdown_to_io=False
        ),
        df[["a", "c"]],
    )

    # Test that all serialisers can deal with categories
    expected = df[["c", "d"]].copy()
    expected["c"] = expected["c"].astype("category")
    # Check that the dtypes match but don't care about the order of the categoricals.
    pdt.assert_frame_equal(
        DataFrameSerializer.restore_dataframe(
            store, key, columns=["c", "d"], categories=["c"]
        ),
        expected,
        check_categorical=False,
    )

    # Test restore w/ empty col list
    pdt.assert_frame_equal(
        DataFrameSerializer.restore_dataframe(store, key, columns=[]), df[[]]
    )


@pytest.mark.parametrize("serialiser", SERLIALISERS)
def test_missing_column(serialiser, store):
    df = pd.DataFrame({"a": [1, 2], "b": [3.0, 4.0], "c": ["∆", "€"], "d": ["#", ";"]})
    key = serialiser.store(store, "prefix", df)

    with pytest.raises(ValueError):
        DataFrameSerializer.restore_dataframe(store, key, columns=["a", "x"])


@pytest.mark.parametrize("serialiser", SERLIALISERS)
def test_dataframe_roundtrip_empty(serialiser, store):
    # Empty column indices vary based on pandas version
    if Version(pd.__version__) < Version("2.0.0.dev0"):
        df = pd.DataFrame({}, columns=pd.Index([], dtype="object"))
    else:
        df = pd.DataFrame({}, columns=pd.Index([], dtype="int64"))
    key = serialiser.store(store, "prefix", df)
    pdt.assert_frame_equal(DataFrameSerializer.restore_dataframe(store, key), df)

    # Test partial restore
    pdt.assert_frame_equal(DataFrameSerializer.restore_dataframe(store, key), df)


@pytest.mark.parametrize("serialiser", SERLIALISERS)
def test_dataframe_roundtrip_no_rows(serialiser, store):
    df = pd.DataFrame({"a": [], "b": [], "c": []}).astype(object)
    key = serialiser.store(store, "prefix", df)
    pdt.assert_frame_equal(DataFrameSerializer.restore_dataframe(store, key), df)

    # Test partial restore
    pdt.assert_frame_equal(
        DataFrameSerializer.restore_dataframe(store, key, columns=["a", "c"]),
        df[["a", "c"]],
    )


def test_filter_query_predicate_exclusion(store):
    with pytest.raises(ValueError):
        DataFrameSerializer.restore_dataframe(
            store, "test.parquet", predicates=[[("a", "==", 1)]], filter_query="True"
        )


def assert_frame_almost_equal(df_left, df_right):
    """Be more friendly to some dtypes that are not preserved during the
    roundtrips."""
    # FIXME: This needs a better documentation
    for col in df_left.columns:
        if pd.api.types.is_datetime64_dtype(
            df_left[col].dtype
        ) and pd.api.types.is_object_dtype(df_right[col].dtype):
            df_right[col] = pd.to_datetime(df_right[col], unit="ns")
        elif pd.api.types.is_object_dtype(
            df_left[col].dtype
        ) and pd.api.types.is_datetime64_dtype(df_right[col].dtype):
            df_left[col] = pd.to_datetime(df_left[col])
        elif (
            len(df_left) > 0
            and pd.api.types.is_object_dtype(df_left[col].dtype)
            and pd.api.types.is_object_dtype(df_right[col].dtype)
        ):
            if isinstance(df_left[col].iloc[0], datetime.date) or isinstance(
                df_right[col].iloc[0], datetime.date
            ):
                df_left[col] = pd.to_datetime(df_left[col], unit="ns")
                df_right[col] = pd.to_datetime(df_right[col], unit="ns")
        elif (
            pd.api.types.is_object_dtype(df_left[col].dtype)
            or pd.api.types.is_string_dtype(df_left[col].dtype)
        ) and isinstance(df_right[col].dtype, pd.CategoricalDtype):
            df_left[col] = df_left[col].astype(df_right[col].dtype)
    pdt.assert_frame_equal(
        df_left.reset_index(drop=True), df_right.reset_index(drop=True)
    )


@pytest.mark.parametrize(
    "df, read_kwargs",
    [
        (pd.DataFrame({"string_ü": ["abc", "affe", "banane", "buchstabe_ü"]}), {}),
        (pd.DataFrame({"integer_ü": np.arange(4)}), {}),
        (pd.DataFrame({"float_ü": [-3.141591, 0.0, 3.141593, 3.141595]}), {}),
        (
            pd.DataFrame(
                {
                    "date_ü": [
                        datetime.date(2011, 1, 31),
                        datetime.date(2011, 2, 3),
                        datetime.date(2011, 2, 4),
                        datetime.date(2011, 3, 10),
                    ]
                }
            ),
            {"date_as_object": False},
        ),
        (
            pd.DataFrame(
                {
                    "date_ü": [
                        datetime.date(2011, 1, 31),
                        datetime.date(2011, 2, 3),
                        datetime.date(2011, 2, 4),
                        datetime.date(2011, 3, 10),
                    ]
                }
            ),
            {"date_as_object": True},
        ),
        (
            pd.DataFrame(
                {"categorical_ü": list("abcd")},
                dtype=pd.api.types.CategoricalDtype(list("abcd"), ordered=True),
            ),
            {},
        ),
    ],
)
@predicate_serialisers
@pytest.mark.parametrize("predicate_pushdown_to_io", [True, False])
def test_predicate_pushdown(
    store, df, read_kwargs, predicate_pushdown_to_io, serialiser
):
    """Test predicate pushdown for several types and operations.

    The DataFrame parameters all need to be of same length for this test
    to work universally. Also the values in the DataFrames need to be
    sorted in ascending order.
    """
    # All test dataframes need to have the same length
    assert len(df) == 4
    assert df[df.columns[0]].is_monotonic_increasing and df.iloc[0, 0] < df.iloc[-1, 0]

    # This is due to the limitation that dates cannot be expressed in
    # Pandas' query() method.
    if isinstance(serialiser, CsvSerializer) and isinstance(
        df.iloc[0, 0], datetime.date
    ):
        pytest.skip("CsvSerialiser cannot filter on dates")

    key = serialiser.store(store, "prefix", df)

    # Test `<` and `>` operators
    expected = df.iloc[[1, 2], :].copy()
    predicates = [
        [(df.columns[0], "<", df.iloc[3, 0]), (df.columns[0], ">", df.iloc[0, 0])]
    ]
    result = serialiser.restore_dataframe(
        store,
        key,
        predicate_pushdown_to_io=predicate_pushdown_to_io,
        predicates=predicates,
        **read_kwargs,
    )
    assert_frame_almost_equal(result, expected)

    # Test `=<` and `>=` operators
    expected = df.iloc[[1, 2, 3], :].copy()
    predicates = [
        [(df.columns[0], "<=", df.iloc[3, 0]), (df.columns[0], ">=", df.iloc[1, 0])]
    ]
    result = serialiser.restore_dataframe(
        store,
        key,
        predicate_pushdown_to_io=predicate_pushdown_to_io,
        predicates=predicates,
        **read_kwargs,
    )
    assert_frame_almost_equal(result, expected)

    # Test `==` operator
    expected = df.iloc[[1], :].copy()
    predicates = [[(df.columns[0], "==", df.iloc[1, 0])]]
    result = serialiser.restore_dataframe(
        store,
        key,
        predicate_pushdown_to_io=predicate_pushdown_to_io,
        predicates=predicates,
        **read_kwargs,
    )
    assert_frame_almost_equal(result, expected)

    # Test `in` operator
    expected = df.iloc[[1], :].copy()
    predicates = [[(df.columns[0], "in", [df.iloc[1, 0]])]]
    result = serialiser.restore_dataframe(
        store,
        key,
        predicate_pushdown_to_io=predicate_pushdown_to_io,
        predicates=predicates,
        **read_kwargs,
    )
    assert_frame_almost_equal(result, expected)

    # Test `!=` operator
    expected = df.iloc[[0, 2, 3], :].copy()
    predicates = [[(df.columns[0], "!=", df.iloc[1, 0])]]
    result = serialiser.restore_dataframe(
        store,
        key,
        predicate_pushdown_to_io=predicate_pushdown_to_io,
        predicates=predicates,
        **read_kwargs,
    )
    assert_frame_almost_equal(result, expected)

    # Test empty DataFrame
    expected = df.head(0)
    predicates = [[(df.columns[0], "<", df.iloc[0, 0])]]
    result = serialiser.restore_dataframe(
        store,
        key,
        predicate_pushdown_to_io=predicate_pushdown_to_io,
        predicates=predicates,
        **read_kwargs,
    )
    assert_frame_almost_equal(result, expected)

    # Test in empty list
    expected = df.head(0)
    predicates = [[(df.columns[0], "in", [])]]
    result = serialiser.restore_dataframe(
        store,
        key,
        predicate_pushdown_to_io=predicate_pushdown_to_io,
        predicates=predicates,
        **read_kwargs,
    )
    assert_frame_almost_equal(result, expected)

    # Test in numpy array
    expected = df.iloc[[1], :].copy()
    predicates = [[(df.columns[0], "in", np.asarray([df.iloc[1, 0], df.iloc[1, 0]]))]]
    result = serialiser.restore_dataframe(
        store,
        key,
        predicate_pushdown_to_io=predicate_pushdown_to_io,
        predicates=predicates,
        **read_kwargs,
    )
    assert_frame_almost_equal(result, expected)

    # Test malformed predicates 1
    predicates = []
    with pytest.raises(ValueError) as exc:
        serialiser.restore_dataframe(
            store,
            key,
            predicate_pushdown_to_io=predicate_pushdown_to_io,
            predicates=predicates,
            **read_kwargs,
        )
    assert str(exc.value) == "Empty predicates"

    # Test malformed predicates 2
    predicates = [[]]
    with pytest.raises(ValueError) as exc:
        serialiser.restore_dataframe(
            store,
            key,
            predicate_pushdown_to_io=predicate_pushdown_to_io,
            predicates=predicates,
            **read_kwargs,
        )
    assert str(exc.value) == "Invalid predicates: Conjunction 0 is empty"

    # Test malformed predicates 3
    predicates = [[(df.columns[0], "<", df.iloc[0, 0])], []]
    with pytest.raises(ValueError) as exc:
        serialiser.restore_dataframe(
            store,
            key,
            predicate_pushdown_to_io=predicate_pushdown_to_io,
            predicates=predicates,
            **read_kwargs,
        )
    assert str(exc.value) == "Invalid predicates: Conjunction 1 is empty"

    # Test malformed predicates 4
    predicates = [[(df.columns[0], "<", df.iloc[0, 0])], ["foo"]]
    with pytest.raises(ValueError) as exc:
        serialiser.restore_dataframe(
            store,
            key,
            predicate_pushdown_to_io=predicate_pushdown_to_io,
            predicates=predicates,
            **read_kwargs,
        )
    assert (
        str(exc.value)
        == "Invalid predicates: Clause 0 in conjunction 1 should be a 3-tuple, got object of type <class 'str'> instead"
    )


@predicate_serialisers
@pytest.mark.parametrize("predicate_pushdown_to_io", [True, False])
def test_predicate_float_equal_big(predicate_pushdown_to_io, store, serialiser):
    df = pd.DataFrame({"float": [3141590.0, 3141592.0, 3141594.0]})
    key = serialiser.store(store, "prefix", df)

    predicates = [[("float", "==", 3141592.0)]]
    result_df = serialiser.restore_dataframe(
        store,
        key,
        predicate_pushdown_to_io=predicate_pushdown_to_io,
        predicates=predicates,
    )
    expected_df = df.iloc[[1], :].copy()

    pdt.assert_frame_equal(
        result_df.reset_index(drop=True), expected_df.reset_index(drop=True)
    )


@predicate_serialisers
@pytest.mark.parametrize("predicate_pushdown_to_io", [True, False])
def test_predicate_float_equal_small(predicate_pushdown_to_io, store, serialiser):
    df = pd.DataFrame({"float": [0.3141590, 0.3141592, 0.3141594]})

    key = serialiser.store(store, "prefix", df)

    predicates = [[("float", "==", 0.3141592)]]
    result_df = serialiser.restore_dataframe(
        store,
        key,
        predicate_pushdown_to_io=predicate_pushdown_to_io,
        predicates=predicates,
    )
    expected_df = df.iloc[[1], :].copy()

    pdt.assert_frame_equal(
        result_df.reset_index(drop=True), expected_df.reset_index(drop=True)
    )


@type_stable_serialisers
@pytest.mark.parametrize("predicate_pushdown_to_io", [True, False])
def test_predicate_eval_string_types(serialiser, store, predicate_pushdown_to_io):
    df = pd.DataFrame({b"a": [1, 2], "b": [3.0, 4.0]})
    key = serialiser.store(store, "prefix", df)
    df.columns = [ensure_unicode_string_type(col) for col in df.columns]
    pdt.assert_frame_equal(DataFrameSerializer.restore_dataframe(store, key), df)

    for col in ["a", b"a", "a"]:
        predicates = [[(col, "==", 1)]]
        result_df = serialiser.restore_dataframe(
            store,
            key,
            predicate_pushdown_to_io=predicate_pushdown_to_io,
            predicates=predicates,
        )

        expected_df = df.iloc[[0], :].copy()

        pdt.assert_frame_equal(
            result_df.reset_index(drop=True), expected_df.reset_index(drop=True)
        )

    for col in ["b", b"b", "b"]:
        predicates = [[(col, "==", 3.0)]]
        result_df = serialiser.restore_dataframe(
            store,
            key,
            predicate_pushdown_to_io=predicate_pushdown_to_io,
            predicates=predicates,
        )

        expected_df = df.iloc[[0], :].copy()

        pdt.assert_frame_equal(
            result_df.reset_index(drop=True), expected_df.reset_index(drop=True)
        )

    for preds in (
        [[("a", "==", 1), ("b", "==", 3.0)]],
        [[("a", "==", 1), (b"b", "==", 3.0)]],
        [[(b"a", "==", 1), ("b", "==", 3.0)]],
    ):
        result_df = serialiser.restore_dataframe(
            store,
            key,
            predicate_pushdown_to_io=predicate_pushdown_to_io,
            predicates=preds,
        )

        expected_df = df.iloc[[0], :].copy()

        pdt.assert_frame_equal(
            result_df.reset_index(drop=True), expected_df.reset_index(drop=True)
        )


@pytest.mark.parametrize(
    "df,value",
    [
        (pd.DataFrame({"u": pd.Series([None], dtype=object)}), "foo"),
        (pd.DataFrame({"b": pd.Series([None], dtype=object)}), b"foo"),
        (pd.DataFrame({"f": pd.Series([np.nan], dtype=float)}), 1.2),
        (
            pd.DataFrame({"t": pd.Series([pd.NaT], dtype="datetime64[ns]")}),
            pd.Timestamp("2017"),
        ),
    ],
)
@predicate_serialisers
@pytest.mark.parametrize("predicate_pushdown_to_io", [True, False])
def test_predicate_pushdown_null_col(
    store, df, value, predicate_pushdown_to_io, serialiser
):
    key = serialiser.store(store, "prefix", df)

    expected = df.iloc[[]].copy()
    predicates = [[(df.columns[0], "==", value)]]
    result = serialiser.restore_dataframe(
        store,
        key,
        predicate_pushdown_to_io=predicate_pushdown_to_io,
        predicates=predicates,
    )
    check_datetimelike_compat = (
        isinstance(value, pd.Timestamp) and not serialiser.type_stable
    )
    pdt.assert_frame_equal(
        result.reset_index(drop=True),
        expected.reset_index(drop=True),
        check_dtype=serialiser.type_stable,
        check_datetimelike_compat=check_datetimelike_compat,
    )


@pytest.mark.parametrize(
    "df, op, value, expected_index",
    [
        (
            pd.DataFrame({"u": pd.Series([None, "x", np.nan], dtype=object)}),
            "==",
            None,
            [0, 2],
        ),
        (
            pd.DataFrame({"u": pd.Series([None, "x", np.nan], dtype=object)}),
            "in",
            [None],
            [0, 2],
        ),
        (
            pd.DataFrame({"u": pd.Series([None, "x", np.nan], dtype=object)}),
            "!=",
            None,
            [1],
        ),
        (
            pd.DataFrame({"u": pd.Series([None, "x", np.nan], dtype=object)}),
            "in",
            [None, "x"],
            [0, 1, 2],
        ),
        (
            pd.DataFrame({"f": pd.Series([np.nan, 1.0, np.nan], dtype=float)}),
            "==",
            np.nan,
            [0, 2],
        ),
        (
            pd.DataFrame({"f": pd.Series([np.nan, 1.0, np.nan], dtype=float)}),
            "in",
            [np.nan],
            [0, 2],
        ),
        (
            pd.DataFrame({"f": pd.Series([np.nan, 1.0, np.nan], dtype=float)}),
            "!=",
            np.nan,
            [1],
        ),
        (
            pd.DataFrame({"f": pd.Series([np.nan, 1.0, np.nan], dtype=float)}),
            "in",
            [np.nan, 1.0],
            [0, 1, 2],
        ),
    ],
)
@predicate_serialisers
@pytest.mark.parametrize("predicate_pushdown_to_io", [True, False])
def test_predicate_parsing_null_values(
    store, df, op, value, expected_index, predicate_pushdown_to_io, serialiser
):
    key = serialiser.store(store, "prefix", df)

    expected = df.iloc[expected_index].copy()
    predicates = [[(df.columns[0], op, value)]]
    result = serialiser.restore_dataframe(
        store,
        key,
        predicate_pushdown_to_io=predicate_pushdown_to_io,
        predicates=predicates,
    )
    if pandas_infer_string() and (not expected.select_dtypes(include=["object"]).empty):
        object_types = expected.select_dtypes(include=["object"]).columns.to_list()
        expected[object_types] = expected[object_types].astype(str)

    assert_frame_equal_non_strict(
        result.reset_index(drop=True),
        expected.reset_index(drop=True),
        check_dtype=serialiser.type_stable,
    )


@pytest.mark.parametrize("op", ["<", "<=", ">", ">="])
@predicate_serialisers
@pytest.mark.parametrize("predicate_pushdown_to_io", [True, False])
def test_predicate_parsing_null_values_failing(
    store, op, predicate_pushdown_to_io, serialiser
):
    df = pd.DataFrame({"u": pd.Series([1.0, np.nan])})
    key = serialiser.store(store, "prefix", df)

    predicates = [[(df.columns[0], op, np.nan)]]
    with pytest.raises(ValueError, match="Only operators supporting null values"):
        serialiser.restore_dataframe(
            store,
            key,
            predicate_pushdown_to_io=predicate_pushdown_to_io,
            predicates=predicates,
        )


@pytest.mark.parametrize(
    "column, expected_null_count",
    [
        ("no_nulls_int", (0, 0, 0)),
        ("partial_nulls_int", (0, 1, 2)),
        ("no_nulls_float", (0, 0, 0)),
        ("partial_nulls_float", (0, 1, 2)),
        ("partial_nulls_obj", (0, 1, 2)),
        ("no_nulls_obj", (0, 0, 0)),
        ("partial_nulls_obj_mixed", (0, 2, 1)),
        ("nulls_reverse_rg", (1, 0, 1)),
    ],
)
def test_null_count(store, column, expected_null_count):
    serialiser = ParquetSerializer(chunk_size=2)

    df = pd.DataFrame(
        {
            "no_nulls_int": [1, 2, 3, 4, 5, 6],
            "partial_nulls_int": [1, 2, 3, None, None, None],
            "no_nulls_float": [1.1, 2.2, 3.3, 4.4, 5.5, 6.6],
            "partial_nulls_float": [1.0, 2.2, 3.3, np.nan, np.nan, np.nan],
            "partial_nulls_obj": [1.0, 2.2, 3.3, np.nan, np.nan, np.nan],
            "no_nulls_obj": ["1.1", "2", "3", "vier", "fuenfeinhalb", "6.6"],
            "partial_nulls_obj_mixed": [1.0, 2.2, None, np.nan, np.nan, 6.6],
            "nulls_reverse_rg": [3.3, np.nan, 1.0, 2.0, np.nan, -1.1],
        }
    )

    key = serialiser.store(store, "prefix", df)
    reader = pa.BufferReader(store.get(key))
    parquet_file = ParquetFile(reader)
    col_idx = parquet_file.reader.column_name_idx(column)

    assert parquet_file.num_row_groups == 3

    for idx in range(0, 3):
        rg = parquet_file.metadata.row_group(idx)
        assert rg.column(col_idx).statistics.null_count == expected_null_count[idx]


@pytest.mark.parametrize(
    "df,value",
    [
        (pd.DataFrame({"nan": pd.Series([np.nan, -1.0, 1.0], dtype=float)}), 0.0),
        (pd.DataFrame({"inf": pd.Series([np.inf, -1.0, 1.0], dtype=float)}), 0.0),
        (pd.DataFrame({"ninf": pd.Series([-np.inf, -1.0, 1.0], dtype=float)}), 0.0),
        (
            pd.DataFrame(
                {"inf2": pd.Series([-np.inf, np.inf, -1.0, 1.0], dtype=float)}
            ),
            0.0,
        ),
        (
            pd.DataFrame(
                {"inf2": pd.Series([-np.inf, np.inf, -1.0, 1.0], dtype=float)}
            ),
            0.0,
        ),
        (
            pd.DataFrame(
                {"inf2": pd.Series([-np.inf, np.inf, -1.0, 1.0], dtype=float)}
            ),
            np.inf,
        ),
        (
            pd.DataFrame(
                {"inf2": pd.Series([-np.inf, np.inf, -1.0, 1.0], dtype=float)}
            ),
            -np.inf,
        ),
    ],
)
@predicate_serialisers
@pytest.mark.parametrize("predicate_pushdown_to_io", [True, False])
def test_predicate_pushdown_weird_floats_col(
    store, df, value, predicate_pushdown_to_io, serialiser
):
    key = serialiser.store(store, "prefix", df)

    col = df.columns[0]

    expected = df.loc[df[col] >= value].copy()
    predicates = [[(col, ">=", value)]]
    result = serialiser.restore_dataframe(
        store,
        key,
        predicate_pushdown_to_io=predicate_pushdown_to_io,
        predicates=predicates,
    )
    assert_frame_almost_equal(result, expected)



"""Functions to get the current time.

plateau modules that need to get the current time should do so only via
this module. This allows test to monkeypatch the methods in this module
to fake certain fixed times.
"""

import datetime
import sys


def datetime_now():
    """Get the current time as datimetime object.

    Same as datetime.datetime.now
    """
    return datetime.datetime.now()


def datetime_utcnow():
    """Get the current time as datimetime object.

    Same as datetime.datetime.utcnow
    """
    if sys.version_info < (3, 11):
        return datetime.datetime.utcnow()
    return datetime.datetime.now(datetime.UTC)



import collections
import inspect


class CopyMixin:
    def copy(self, **kwargs):
        constructor_args = inspect.signature(self.__init__).parameters  # type: ignore
        init_args = collections.OrderedDict()

        for arg in constructor_args.keys():
            if arg == "self":
                continue
            init_args[arg] = kwargs.get(arg, getattr(self, arg, None))

        return type(self)(**init_args)



import pickle

import dask.bag as db
import pytest

from plateau.io.dask.bag import store_bag_as_dataset
from plateau.io.testing.write import *  # noqa


def _store_dataframes(df_list, *args, **kwargs):
    bag = store_bag_as_dataset(db.from_sequence(df_list), *args, **kwargs)
    s = pickle.dumps(bag, pickle.HIGHEST_PROTOCOL)
    bag = pickle.loads(s)
    return bag.compute()


@pytest.fixture()
def bound_store_dataframes(request):
    return _store_dataframes



from functools import partial

import pytest

from plateau.io.iter import (
    read_dataset_as_dataframes__iterator,
    read_dataset_as_metapartitions__iterator,
)
from plateau.io.testing.read import *  # noqa


@pytest.fixture(params=["dataframe"])
def output_type(request):
    return request.param


def _load_dataframes(output_type, *args, **kwargs):
    if output_type == "dataframe":
        func = read_dataset_as_dataframes__iterator
    else:
        raise ValueError(f"Unknown output type {output_type}")
    return list(func(*args, **kwargs))


@pytest.fixture()
def bound_load_dataframes(output_type):
    return partial(_load_dataframes, output_type)


def _load_metapartitions(*args, **kwargs):
    return list(read_dataset_as_metapartitions__iterator(*args, **kwargs))


@pytest.fixture()
def bound_load_metapartitions():
    return _load_metapartitions



import numpy as np
import pandas as pd
import pyarrow as pa
import pytest

from plateau.io.eager import (
    commit_dataset,
    create_empty_dataset_header,
    update_dataset_from_dataframes,
    write_single_partition,
)
from plateau.io.testing.update import *  # noqa: F4, F403


@pytest.fixture()
def bound_update_dataset():
    return update_dataset_from_dataframes


def test_create_empty_header_from_pyarrow_schema(store_factory):
    # GH228
    df = pd.DataFrame(
        [{"part": 1, "id": 1, "col1": "abc"}, {"part": 2, "id": 2, "col1": np.nan}]
    )
    dataset_uuid = "sample_ds"
    schema = pa.Schema.from_pandas(df)

    dm = create_empty_dataset_header(
        store=store_factory,
        dataset_uuid=dataset_uuid,
        schema=schema,
        partition_on=["part"],
    )

    new_partitions = [
        write_single_partition(
            store=store_factory,
            dataset_uuid=dataset_uuid,
            data=[df.loc[df["part"] == 1]],
            partition_on=["part"],
        )
    ]
    assert len(dm.partitions) == 0
    dm = commit_dataset(
        store=store_factory,
        dataset_uuid=dataset_uuid,
        new_partitions=new_partitions,
        partition_on=["part"],
    )

    assert len(dm.partitions) == 1



from plateau.core import naming
from plateau.core.naming import metadata_key_from_uuid


def delete_common_metadata(dataset_factory):
    for table in dataset_factory.tables:
        key = f"{dataset_factory.uuid}/{table}/{naming.TABLE_METADATA_FILE}"
        dataset_factory.store.delete(key)
    return dataset_factory


def delete_indices(dataset_factory):
    for index_object in dataset_factory.indices.values():
        index_key = index_object.index_storage_key
        dataset_factory.store.delete(index_key)
    return dataset_factory


def delete_top_level_metadata(dataset_factory, *args):
    """The additional arguments allow to schedule this function with delayed
    objects."""
    dataset_factory.store.delete(metadata_key_from_uuid(dataset_factory.dataset_uuid))



import pytest

from plateau.io.iter import store_dataframes_as_dataset__iter
from plateau.io.testing.write import *  # noqa


def _store_dataframes(df_list, *args, **kwargs):
    df_generator = (x for x in df_list)
    return store_dataframes_as_dataset__iter(df_generator, *args, **kwargs)


@pytest.fixture()
def bound_store_dataframes():
    return _store_dataframes



# pylint: disable=E1101


from collections import OrderedDict
from functools import partial
from typing import cast

import numpy as np
import pandas as pd
import pandas.testing as pdt
import pytest
from minimalkv import get_store_from_url

from plateau.core.dataset import DatasetMetadata
from plateau.core.index import ExplicitSecondaryIndex
from plateau.core.uuid import gen_uuid
from plateau.io.eager import read_table
from plateau.io_components.metapartition import MetaPartition
from plateau.serialization import DataFrameSerializer


class NoPickle:
    def __getstate__(self):
        raise RuntimeError("do NOT pickle this object!")


def mark_nopickle(obj):
    obj._nopickle = NoPickle()


def no_pickle_store(url):
    store = get_store_from_url(url)
    mark_nopickle(store)
    return store


def no_pickle_factory(url):
    return partial(no_pickle_store, url)


@pytest.fixture(params=["URL", "KeyValue", "Callable"])
def store_input_types(request, tmpdir):
    url = f"hfs://{tmpdir}"

    if request.param == "URL":
        return url
    elif request.param == "KeyValue":
        return get_store_from_url(url)
    elif request.param == "Callable":
        return no_pickle_factory(url)
    else:
        raise RuntimeError(f"Encountered unknown store type {type(request.param)}")


def test_store_input_types(store_input_types, bound_store_dataframes):
    from plateau.serialization.testing import get_dataframe_not_nested

    dataset_uuid = "dataset_uuid"
    df = get_dataframe_not_nested(10)

    assert bound_store_dataframes(
        [df],
        dataset_uuid=dataset_uuid,
        store=store_input_types,
        partition_on=[df.columns[0]],
        secondary_indices=[df.columns[1]],
    )


def test_file_structure_dataset_v4(store_factory, bound_store_dataframes):
    df = pd.DataFrame(
        {"P": np.arange(0, 10), "L": np.arange(0, 10), "TARGET": np.arange(10, 20)}
    )

    df_list = [df.copy(deep=True), df.copy(deep=True)]

    dataset = bound_store_dataframes(
        df_list, store=store_factory, dataset_uuid="dataset_uuid", metadata_version=4
    )

    assert isinstance(dataset, DatasetMetadata)
    assert len(dataset.partitions) == 2

    store = store_factory()

    assert len(store.keys()) == 4
    assert "dataset_uuid/table/_common_metadata" in store
    assert "dataset_uuid.by-dataset-metadata.json" in store


def test_file_structure_dataset_v4_partition_on(store_factory, bound_store_dataframes):
    store = store_factory()
    assert set(store.keys()) == set()
    df = pd.DataFrame(
        {"P": [1, 2, 3, 1, 2, 3], "L": [1, 1, 1, 2, 2, 2], "TARGET": np.arange(10, 16)}
    )

    df_list = [df.copy(deep=True), df.copy(deep=True)]
    dataset = bound_store_dataframes(
        df_list,
        store=store_factory,
        dataset_uuid="dataset_uuid",
        partition_on=["P", "L"],
        metadata_version=4,
    )

    assert isinstance(dataset, DatasetMetadata)

    assert dataset.partition_keys == ["P", "L"]

    assert len(dataset.partitions) == 12

    store = store_factory()
    actual_keys = set(store.keys())
    assert len(actual_keys) == 14  # one per partition + json + schema


def test_store_dataframes_as_dataset(
    store_factory, metadata_version, bound_store_dataframes
):
    df = pd.DataFrame(
        {"P": np.arange(0, 10), "L": np.arange(0, 10), "TARGET": np.arange(10, 20)}
    )

    df_list = [df.copy(deep=True), df.copy(deep=True)]

    dataset = bound_store_dataframes(
        df_list,
        store=store_factory,
        dataset_uuid="dataset_uuid",
        metadata_version=metadata_version,
        secondary_indices=["P"],
    )

    assert isinstance(dataset, DatasetMetadata)
    assert len(dataset.partitions) == 2

    assert "P" in dataset.indices

    store = store_factory()
    stored_dataset = DatasetMetadata.load_from_store("dataset_uuid", store)
    assert dataset.uuid == stored_dataset.uuid
    assert dataset.metadata == stored_dataset.metadata
    assert dataset.partitions == stored_dataset.partitions

    index_dct = (
        cast(ExplicitSecondaryIndex, stored_dataset.indices["P"]).load(store).index_dct
    )
    assert sorted(index_dct.keys()) == list(range(0, 10))

    counter = 0
    for k in store.keys():
        if "parquet" in k and "indices" not in k:
            counter += 1
            df_stored = DataFrameSerializer.restore_dataframe(key=k, store=store)
            pdt.assert_frame_equal(df, df_stored)
    assert counter == 2


def test_store_dataframes_as_dataset_empty_dataframe(
    store_factory, metadata_version, df_all_types, bound_store_dataframes
):
    """Test that writing an empty column succeeds.

    In particular, this may fail due to too strict schema validation.
    """
    df_empty = df_all_types.drop(0)

    assert df_empty.empty
    df_list = [df_empty]

    dataset = bound_store_dataframes(
        df_list,
        store=store_factory,
        dataset_uuid="dataset_uuid",
        metadata_version=metadata_version,
    )

    assert isinstance(dataset, DatasetMetadata)
    assert len(dataset.partitions) == 1

    store = store_factory()
    stored_dataset = DatasetMetadata.load_from_store("dataset_uuid", store)
    assert dataset.uuid == stored_dataset.uuid
    assert dataset.metadata == stored_dataset.metadata
    assert dataset.partitions == stored_dataset.partitions

    df_stored = DataFrameSerializer.restore_dataframe(
        key=next(iter(dataset.partitions.values())).files["table"], store=store
    )
    pdt.assert_frame_equal(df_empty, df_stored)


def test_store_dataframes_as_dataset_batch_mode(
    store_factory, metadata_version, bound_store_dataframes
):
    # TODO: Kick this out?
    values_p1 = [1, 2, 3]
    values_p2 = [4, 5, 6]
    df = pd.DataFrame({"P": values_p1})
    df2 = pd.DataFrame({"P": values_p2})

    df_list = [[df, df2]]

    dataset = bound_store_dataframes(
        df_list,
        store=store_factory,
        dataset_uuid="dataset_uuid",
        metadata_version=metadata_version,
        secondary_indices="P",
    )

    assert isinstance(dataset, DatasetMetadata)
    assert len(dataset.partitions) == 2

    store = store_factory()
    stored_dataset = DatasetMetadata.load_from_store(
        "dataset_uuid", store
    ).load_all_indices(store)
    assert dataset.uuid == stored_dataset.uuid
    assert dataset.metadata == stored_dataset.metadata
    assert dataset.partitions == stored_dataset.partitions

    assert "P" in dataset.indices


def test_store_dataframes_as_dataset_auto_uuid(
    store_factory, metadata_version, mock_uuid, bound_store_dataframes
):
    df = pd.DataFrame(
        {"P": np.arange(0, 10), "L": np.arange(0, 10), "TARGET": np.arange(10, 20)}
    )

    df_list = [df.copy(deep=True)]

    dataset = bound_store_dataframes(
        df_list, store=store_factory, metadata_version=metadata_version
    )

    assert isinstance(dataset, DatasetMetadata)
    assert len(dataset.partitions) == 1

    stored_dataset = DatasetMetadata.load_from_store(
        "auto_dataset_uuid", store_factory()
    )
    assert dataset.uuid == stored_dataset.uuid
    assert dataset.metadata == stored_dataset.metadata
    assert dataset.partitions == stored_dataset.partitions


def test_store_dataframes_as_dataset_mp_partition_on_none(
    metadata_version, store, store_factory, bound_store_dataframes
):
    df = pd.DataFrame(
        {"P": np.arange(0, 10), "L": np.arange(0, 10), "TARGET": np.arange(10, 20)}
    )

    mp = MetaPartition(label=gen_uuid(), data=df, metadata_version=metadata_version)

    df_list = [None, mp]
    dataset = bound_store_dataframes(
        df_list,
        store=store_factory,
        dataset_uuid="dataset_uuid",
        metadata_version=metadata_version,
        partition_on=["P"],
    )

    assert isinstance(dataset, DatasetMetadata)
    assert dataset.partition_keys == ["P"]
    assert len(dataset.partitions) == 10
    assert dataset.metadata_version == metadata_version

    stored_dataset = DatasetMetadata.load_from_store("dataset_uuid", store)

    assert dataset == stored_dataset


def test_store_dataframes_partition_on(store_factory, bound_store_dataframes):
    df = pd.DataFrame(
        OrderedDict([("location", ["0", "1", "2"]), ("other", ["a", "a", "a"])])
    )

    # First partition is empty, test this edgecase
    input_ = [df.head(0), df]
    dataset = bound_store_dataframes(
        input_,
        store=store_factory,
        dataset_uuid="dataset_uuid",
        metadata_version=4,
        partition_on=["other"],
        secondary_indices="location",
    )

    assert len(dataset.partitions) == 1
    assert len(dataset.indices) == 1

    assert dataset.partition_keys == ["other"]


def _exception_str(exception):
    """Extract the exception message, even if this is a re-throw of an
    exception in distributed."""
    if isinstance(exception, ValueError) and exception.args[0] == "Long error message":
        return exception.args[1]
    return str(exception)


@pytest.mark.parametrize(
    "dfs,ok",
    [
        (
            [
                pd.DataFrame(
                    {
                        "P": pd.Series([1], dtype=np.int64),
                        "X": pd.Series([1], dtype=np.int64),
                    }
                ),
                pd.DataFrame(
                    {
                        "P": pd.Series([2], dtype=np.int64),
                        "X": pd.Series([2], dtype=np.int64),
                    }
                ),
            ],
            True,
        ),
        (
            [
                pd.DataFrame(
                    {
                        "P": pd.Series([1], dtype=np.int64),
                        "X": pd.Series([1], dtype=np.int32),
                    }
                ),
                pd.DataFrame(
                    {
                        "P": pd.Series([2], dtype=np.int64),
                        "X": pd.Series([2], dtype=np.int16),
                    }
                ),
            ],
            True,
        ),
        (
            [
                pd.DataFrame(
                    {
                        "P": pd.Series([1], dtype=np.int16),
                        "X": pd.Series([1], dtype=np.int64),
                    }
                ),
                pd.DataFrame(
                    {
                        "P": pd.Series([2], dtype=np.int32),
                        "X": pd.Series([2], dtype=np.int64),
                    }
                ),
            ],
            True,
        ),
        (
            [
                pd.DataFrame(
                    {
                        "P": pd.Series([1], dtype=np.int64),
                        "X": pd.Series([1], dtype=np.int64),
                    }
                ),
                pd.DataFrame(
                    {
                        "P": pd.Series([2], dtype=np.int64),
                        "X": pd.Series([2], dtype=np.uint64),
                    }
                ),
            ],
            False,
        ),
        (
            [
                pd.DataFrame(
                    {
                        "P": pd.Series([1], dtype=np.int64),
                        "X": pd.Series([1], dtype=np.int64),
                    }
                ),
                pd.DataFrame(
                    {
                        "P": pd.Series([2], dtype=np.int64),
                        "X": pd.Series([2], dtype=np.int64),
                        "Y": pd.Series([2], dtype=np.int64),
                    }
                ),
            ],
            False,
        ),
        (
            [
                pd.DataFrame(
                    {
                        "P": pd.Series([1, 2], dtype=np.int64),
                        "X": pd.Series([1, 2], dtype=np.int64),
                    }
                ),
                pd.DataFrame(
                    {
                        "P": pd.Series([3], dtype=np.int64),
                        "X": pd.Series([3], dtype=np.uint64),
                    }
                ),
            ],
            False,
        ),
    ],
)
def test_schema_check_write(dfs, ok, store_factory, bound_store_dataframes):
    if ok:
        bound_store_dataframes(
            dfs,
            store=store_factory,
            dataset_uuid="dataset_uuid",
            partition_on=["P"],
            metadata_version=4,
        )
    else:
        with pytest.raises(Exception) as exc:
            bound_store_dataframes(
                dfs,
                store=store_factory,
                dataset_uuid="dataset_uuid",
                partition_on=["P"],
                metadata_version=4,
            )
        assert (
            "Schemas for dataset 'dataset_uuid' are not compatible!"
            in _exception_str(exc.value)
        )


@pytest.mark.xfail(reason="mocking doesn't work for dask atm")
def test_schema_check_write_nice_error(
    store_factory, bound_store_dataframes, mock_uuid
):
    df1 = pd.DataFrame(
        {
            "P": pd.Series([1, 1], dtype=np.int64),
            "Q": pd.Series([1, 2], dtype=np.int64),
            "X": pd.Series([1, 1], dtype=np.int64),
        }
    )
    df2 = pd.DataFrame(
        {
            "P": pd.Series([2, 2], dtype=np.uint64),
            "Q": pd.Series([1, 2], dtype=np.int64),
            "X": pd.Series([1, 1], dtype=np.int64),
        }
    )
    df_list = [
        df1,
        df2,
    ]
    with pytest.raises(Exception) as exc:
        bound_store_dataframes(
            df_list,
            store=store_factory,
            dataset_uuid="dataset_uuid",
            partition_on=["P", "Q"],
            metadata_version=4,
        )

    assert _exception_str(exc.value).startswith(
        """Schemas for dataset 'dataset_uuid' are not compatible!

Schema violation

Origin schema: {P=2/Q=2/auto_dataset_uuid}
Origin reference: {P=1/Q=2/auto_dataset_uuid}

Diff:
"""
    )


@pytest.mark.xfail(reason="mocking doesn't work for dask atm")
def test_schema_check_write_cut_error(store_factory, bound_store_dataframes, mock_uuid):
    df1 = pd.DataFrame(
        {
            "P": pd.Series([1] * 100, dtype=np.int64),
            "Q": pd.Series(range(100), dtype=np.int64),
            "X": pd.Series([1] * 100, dtype=np.int64),
        }
    )
    df2 = pd.DataFrame(
        {
            "P": pd.Series([2] * 100, dtype=np.uint64),
            "Q": pd.Series(range(100), dtype=np.int64),
            "X": pd.Series([1] * 100, dtype=np.int64),
        }
    )
    df_list = [
        df1,
        df2,
    ]
    with pytest.raises(Exception) as exc:
        bound_store_dataframes(
            df_list,
            store=store_factory,
            dataset_uuid="dataset_uuid",
            partition_on=["P", "Q"],
            metadata_version=4,
        )
    assert _exception_str(exc.value).startswith(
        """Schemas for dataset 'dataset_uuid' are not compatible!

Schema violation

Origin schema: {P=2/Q=99/auto_dataset_uuid}
Origin reference: {P=1/Q=99/auto_dataset_uuid}

Diff:
"""
    )


def test_metadata_consistency_errors_fails(
    store_factory, metadata_version, bound_store_dataframes
):
    df = pd.DataFrame({"W": np.arange(0, 10), "L": np.arange(0, 10)})

    df_2 = pd.DataFrame(
        {"P": np.arange(10, 20), "L": np.arange(0, 10), "TARGET": np.arange(10, 20)}
    )

    df_list = [df, df_2]

    # Also test `df_list` in reverse order, as this could lead to different results
    for dfs in [df_list, list(reversed(df_list))]:
        with pytest.raises(
            Exception, match=r"Schemas for dataset .* are not compatible!"
        ):
            return bound_store_dataframes(
                dfs, store=store_factory, metadata_version=metadata_version
            )


def test_store_dataframes_as_dataset_overwrite(
    store_factory, dataset_function, bound_store_dataframes
):
    with pytest.raises(RuntimeError):
        bound_store_dataframes(
            [pd.DataFrame()], store=store_factory, dataset_uuid=dataset_function.uuid
        )
    bound_store_dataframes(
        [pd.DataFrame()],
        store=store_factory,
        dataset_uuid=dataset_function.uuid,
        overwrite=True,
    )
    bound_store_dataframes(
        [pd.DataFrame()], store=store_factory, dataset_uuid="new_dataset_uuid"
    )


@pytest.mark.skip("What is the intended behaviour for this?")
def test_store_empty_dataframes_partition_on(store_factory, bound_store_dataframes):
    df1 = pd.DataFrame({"x": [1], "y": [1]}).iloc[[]]
    md1 = bound_store_dataframes(
        [df1], store=store_factory, dataset_uuid="uuid", partition_on=["x"]
    )
    assert md1.tables == ["table"]
    assert set(md1.schema.names) == set(df1.columns)

    df2 = pd.DataFrame({"x": [1], "y": [1], "z": [1]}).iloc[[]]
    md2 = bound_store_dataframes(
        [df2],
        store=store_factory,
        dataset_uuid="uuid",
        partition_on=["x"],
        overwrite=True,
    )
    assert md2.tables == ["table"]
    assert set(md2.schema.names) == set(df2.columns)

    df3 = pd.DataFrame({"x": [1], "y": [1], "a": [1]}).iloc[[]]
    md3 = bound_store_dataframes(
        [{"table2": df3}],
        store=store_factory,
        dataset_uuid="uuid",
        partition_on=["x"],
        overwrite=True,
    )
    assert md3.tables == ["table2"]
    assert set(md3.schema.names) == set(df3.columns)


@pytest.mark.skip("What is the intended behaviour for this?")
def test_store_overwrite_none(store_factory, bound_store_dataframes):
    df1 = pd.DataFrame({"x": [1], "y": [1]})
    md1 = bound_store_dataframes(
        [df1], store=store_factory, dataset_uuid="uuid", partition_on=["x"]
    )
    assert md1.tables == ["table"]
    assert set(md1.schema.names) == set(df1.columns)

    md2 = bound_store_dataframes(
        [None],
        store=store_factory,
        dataset_uuid="uuid",
        partition_on=["x"],
        overwrite=True,
    )
    assert md2.tables == []


def test_secondary_index_on_partition_column(store_factory, bound_store_dataframes):
    df1 = pd.DataFrame({"x": [1], "y": [1]})
    with pytest.raises(
        RuntimeError, match="Cannot create secondary index on partition columns: {'x'}"
    ):
        bound_store_dataframes(
            [df1], store=store_factory, partition_on=["x"], secondary_indices=["x"]
        )


def test_non_default_table_name_roundtrip(store_factory, bound_store_dataframes):
    df = pd.DataFrame({"A": [1]})
    bound_store_dataframes(
        [df], store=store_factory, dataset_uuid="dataset_uuid", table_name="foo"
    )
    for k in store_factory():
        if k.endswith(".parquet") and "indices" not in k:
            assert "foo" in k
    result = read_table(dataset_uuid="dataset_uuid", store=store_factory)

    pdt.assert_frame_equal(df, result)
